#  coding: utf-8
# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref entity's classes and adapters for SEDA classes"""

from lxml import etree

from cubicweb.uilib import cut
from cubicweb.entities import AnyEntity, fetch_config
from cubicweb.predicates import is_instance, is_in_state

from cubes.compound.entities import IClonableAdapter
from cubes.oaipmh.entities import OAISetSpec

from cubes.saem_ref import cwuri_url
from cubes.saem_ref.entities import AbstractXmlAdapter, oai

_ = unicode


class SEDAComponentIClonableAdapter(IClonableAdapter):
    """Cloning adapter for SEDA components."""
    rtype = 'seda_clone_of'
    skiprtypes = ()


class SEDAProfileIClonableAdapter(IClonableAdapter):
    """Cloning adapter for SEDA profiles."""
    __select__ = IClonableAdapter.__select__ & is_in_state('published')
    rtype = 'seda_replace'
    skiprtypes = ()


class SEDAProfile(AnyEntity):
    __regid__ = 'SEDAProfile'
    fetch_attrs, cw_fetch_order = fetch_config(('title', 'ark',
                                                'description', 'description_format'))

    @property
    def archives(self):
        return self.reverse_seda_parent

    def dc_title(self):
        return self.title or self.ark

    def __getstate__(self):
        # Exclude ark from copy to have so that it can be generated for the clone
        odict = self.__dict__.copy()
        odict['cw_attr_cache'].pop('ark', None)
        return odict

    @property
    def predecessor(self):
        """The predecessor of the current profile, that is the one that has been replaced by it"""
        predecessors = self.seda_replace
        if predecessors:
            return predecessors[0]
        return None

    @property
    def successor(self):
        """The successor of the current profile, that is the one that has replaced it"""
        successors = self.reverse_seda_replace
        if successors:
            return successors[0]
        return None


class TransferringAgentOAISetSpec(OAISetSpec):
    """OAI-PMH set specifier to match SEDAProfile related to a transferring
    agent.
    """

    def setspecs(self, cnx):
        rset = cnx.execute(
            'Any A WHERE X archival_role R, R name "deposit", X ark A')
        desc = 'SEDAProfile with transferring agent identified by {0}'
        for value, in rset.rows:
            yield value, desc.format(value)

    def setspec_restrictions(self, value):
        qs = ('X seda_transferring_agent A, A archival_role R'
              ', R name "deposit", A ark %(value)s'
              ', X in_state ST, NOT ST name "draft"')
        return qs, {'value': value}


class SEDAProfileOAIPMHRecordAdapter(oai.OAIPMHActiveRecordAdapter):
    """OAI-PMH adapter for SEDAProfile entity type."""
    __select__ = oai.OAIPMHActiveRecordAdapter.__select__ & is_instance('SEDAProfile')
    metadata_view = 'saem_ref.seda02'

    @classmethod
    def set_definition(cls):
        specifier = oai.PublicETypeOAISetSpec(
            'SEDAProfile', cls.identifier_attribute, 'profile')
        specifier['transferring_agent'] = TransferringAgentOAISetSpec()
        return specifier


class ProfileArchiveObject(AnyEntity):
    __regid__ = 'ProfileArchiveObject'
    fetch_attrs, cw_fetch_order = fetch_config(('seda_name',))

    def dc_title(self):
        """Return the dc_title for an entity supporting seda_name relation."""
        if self.name.value:
            arch_title = u'({0})'.format(self.name.value)
        else:
            arch_title = u'#{0}'.format(self.eid)
        return u'{0} {1}'.format(self.dc_type(), arch_title)

    @property
    def name(self):
        return self.seda_name[0]

    @property
    def content_description(self):
        if self.seda_content_description:
            return self.seda_content_description[0]
        return None

    @property
    def access_restriction_code(self):
        if self.seda_access_restriction_code:
            return self.seda_access_restriction_code[0]
        return None

    @property
    def appraisal_rule(self):
        if self.seda_appraisal_rule:
            return self.seda_appraisal_rule[0]
        return None

    @property
    def first_level(self):
        """True if this document unit's parent is a profile"""
        return bool(self.seda_parent) and self.seda_parent[0].cw_etype == 'SEDAProfile'


class ProfileDocument(AnyEntity):
    __regid__ = 'ProfileDocument'
    fetch_attrs, cw_fetch_order = fetch_config(('user_annotation', 'seda_document_type_code'))

    @property
    def description(self):
        if self.seda_description:
            return self.seda_description[0]
        return None

    @property
    def character_set_code(self):
        if self.seda_character_set_code:
            return self.seda_character_set_code[0]
        return None

    @property
    def file_type_code(self):
        if self.seda_file_type_code:
            return self.seda_file_type_code[0]
        return None

    @property
    def document_type_code(self):
        return self.seda_document_type_code[0]

    def dc_title(self):
        dtype = self.document_type_code.seda_document_type_code_value[0].dc_title()
        if self.user_annotation:
            title = u'%s (%s)' % (cut(self.user_annotation, 100), dtype)
        else:
            title = u'%s #%s' % (dtype, self.eid)
        return title


def _seda_profile_element_dc_title(entity):
    """Return a dc_title for an "SEDA profile element"-like entity."""
    return u'{0} #{1}'.format(entity.dc_type(), entity.eid)


class SEDAContentDescription(AnyEntity):
    __regid__ = 'SEDAContentDescription'
    fetch_attrs, cw_fetch_order = fetch_config(('seda_description_level',))

    @property
    def description(self):
        if self.seda_description:
            return self.seda_description[0]
        return None

    @property
    def description_level(self):
        return self.seda_description_level[0]

    @property
    def keywords(self):
        return self.reverse_seda_keyword_of

    def dc_title(self):
        return _seda_profile_element_dc_title(self)


class SEDADescriptionLevel(AnyEntity):
    __regid__ = 'SEDADescriptionLevel'

    def dc_title(self):
        return _seda_profile_element_dc_title(self)


class SEDAAppraisalRule(AnyEntity):
    __regid__ = 'SEDAAppraisalRule'
    fetch_attrs, cw_fetch_order = fetch_config(('seda_appraisal_rule_code',
                                                'seda_appraisal_rule_duration'))

    def dc_title(self):
        return _seda_profile_element_dc_title(self)

    @property
    def appraisal_rule_code(self):
        return self.seda_appraisal_rule_code[0]

    @property
    def appraisal_rule_duration(self):
        return self.seda_appraisal_rule_duration[0]


class SEDAAccessRestrictionCode(AnyEntity):
    __regid__ = 'SEDAAccessRestrictionCode'

    def dc_title(self):
        return _seda_profile_element_dc_title(self)


class SEDAName(AnyEntity):
    __regid__ = 'SEDAName'
    fetch_attrs, cw_fetch_order = fetch_config(('value',))

    def dc_title(self):
        return self.printable_value('value') or self._cw._('<no value>')


class SEDADescription(AnyEntity):
    __regid__ = 'SEDADescription'
    fetch_attrs, cw_fetch_order = fetch_config(('value',))

    def dc_title(self):
        return self.printable_value('value') or self._cw._('<no value>')


class SEDADate(AnyEntity):
    __regid__ = 'SEDADate'
    fetch_attrs, cw_fetch_order = fetch_config(('value',))

    def dc_title(self):
        return self.printable_value('value') or self._cw._('<no value>')


class SEDAKeyword(AnyEntity):
    __regid__ = 'SEDAKeyword'

    @property
    def scheme(self):
        return self.seda_keyword_scheme[0]

    @property
    def concept(self):
        if self.seda_keyword_value:
            return self.seda_keyword_value[0]
        return None

    def dc_title(self):
        if self.concept:
            return self.concept.label()
        if self.scheme:
            return self._cw._('value from scheme "%s"') % self.scheme.dc_title()
        return self._cw._('<no value>')


# XSD transformation ###############################################################################

class XSDAttr(dict):
    """Simple object to define an xsd:attribute element.

    Params:

    * `name`, value for the 'name' attribute of the xsd:attribute

    * `type`, value for the 'type' attribute of the xsd:attribute

    * `cardinality`, optional cardinality that will be used to compute
      the 'use' attribute of the xsd:attribute (default to 'prohibited')

    * `value`, optional value for the 'fixed' attribute of the xsd:attribute
    """
    def __init__(self, name, xs_type, cardinality=None, value=None):
        assert cardinality in (None, '1', '0..1'), cardinality
        if cardinality is None:
            use = 'prohibited'
        elif cardinality == '1':
            use = 'required'
        else:
            use = 'optional'
        super(XSDAttr, self).__init__(name=name, type=xs_type, use=use)
        if value is not None:
            self['fixed'] = unicode(value)


LIST_VERSION_ID_2009 = XSDAttr('listVersionID', 'xsd:token', '1', 'edition 2009')
LIST_VERSION_ID_2011 = XSDAttr('listVersionID', 'xsd:token', '1', 'edition 2011')


class SEDA1XSDExport(AbstractXmlAdapter):
    """Adapter to build an XSD representation of a SEDA profile, using SEDA 1.0 specification"""
    __regid__ = 'SEDA-1.0.xsd'
    __select__ = is_instance('SEDAProfile')

    namespaces = {
        None: 'fr:gouv:culture:archivesdefrance:seda:v1.0',
        'xsd': 'http://www.w3.org/2001/XMLSchema',
        'qdt': 'fr:gouv:culture:archivesdefrance:seda:v1.0:QualifiedDataType:1',
        'udt': 'urn:un:unece:uncefact:data:standard:UnqualifiedDataType:10',
        'clmDAFFileTypeCode': 'urn:un:unece:uncefact:codelist:draft:DAF:fileTypeCode:2009-08-18',
        'clmIANACharacterSetCode':
        'urn:un:unece:uncefact:codelist:standard:IANA:CharacterSetCode:2007-05-14',
        'clmIANAMIMEMediaType':
        'urn:un:unece:uncefact:codelist:standard:IANA:MIMEMediaType:2008-11-12',
        'clm60133': 'urn:un:unece:uncefact:codelist:standard:6:0133:40106',
    }
    root_attributes = {
        'attributeFormDefault': 'unqualified',
        'elementFormDefault': 'qualified',
        'targetNamespace': 'fr:gouv:culture:archivesdefrance:seda:v1.0',
        'version': '1.0',
    }

    @property
    def file_name(self):
        """Return a file name for the dump"""
        return '%s.xsd' % self.entity.dc_title()

    def dump(self):
        """Return an XSD string for the adapted SEDA profile."""
        root = self.element('xsd:schema', attributes=self.root_attributes)
        self.xsd_transfer(root, self.entity)
        return etree.tostring(root, encoding=self.encoding, pretty_print=True, standalone=False)

    # business visit methods #######################################################################

    def xsd_transfer(self, parent, profile):
        """Append XSD elements for the profile to the given parent node."""
        transfer_node = self.xsd_element(parent, 'ArchiveTransfer',
                                         annotation=profile.title,
                                         xsd_attributes=[XSDAttr('Id', 'xsd:ID')])
        if profile.description:
            self.xsd_element(transfer_node, 'Comment', 'udt:TextType',
                             value=profile.printable_value('description', format='text/plain'),
                             xsd_attributes=[XSDAttr('languageID', 'xsd:language')])
        self.xsd_element(transfer_node, 'Date', 'udt:DateTimeType')
        self.xsd_element(transfer_node, 'TransferIdentifier', 'qdt:ArchivesIDType',
                         xsd_attributes=self.xsd_attributes_scheme())
        for archive in profile.archives:
            self.xsd_archive(transfer_node, archive)

    def xsd_agent(self, parent, agent, role, agent_type='OrganizationType', name_model='Name'):
        """Append XSD elements for the transferring/archival agency to the given parent node."""
        agent_node = self.xsd_element(parent, role, agent_type)
        self.xsd_element(agent_node, 'Identification', 'qdt:ArchivesIDType',
                         value=agent.ark,
                         xsd_attributes=self.xsd_attributes_scheme())
        self.xsd_element(agent_node, name_model, 'udt:TextType', agent.name,
                         xsd_attributes=[XSDAttr('languageID', 'xsd:language')])
        return agent_node

    def xsd_archive(self, parent, archive):
        """Append XSD elements for an archive to the given parent node."""
        archive_node = self.xsd_element(parent, 'Archive',
                                        cardinality=archive.user_cardinality,
                                        annotation=archive.user_annotation,
                                        xsd_attributes=[XSDAttr('Id', 'xsd:ID')])
        # hard-coded description's language
        self.xsd_element(archive_node, 'DescriptionLanguage', 'qdt:CodeLanguageType',
                         value='fra',
                         xsd_attributes=[LIST_VERSION_ID_2011])
        self.xsd_element(archive_node, 'Name', 'udt:TextType',
                         value=archive.name.value,
                         annotation=archive.name.user_annotation,
                         xsd_attributes=[XSDAttr('languageID', 'xsd:language')])
        if archive.appraisal_rule:
            self.xsd_appraisal_rule(archive_node, archive.appraisal_rule)
        self.xsd_access_restriction(archive_node, archive.access_restriction_code)
        self.xsd_content_description(archive_node, archive.content_description)
        self.xsd_children(archive_node, archive)

    archive_object_tag_name = 'ArchiveObject'

    def xsd_archive_object(self, parent, archive_object):
        """Append XSD elements for the archive object to the given parent node."""
        ao_node = self.xsd_element(parent, self.archive_object_tag_name,
                                   cardinality=archive_object.user_cardinality,
                                   annotation=archive_object.user_annotation,
                                   xsd_attributes=[XSDAttr('Id', 'xsd:ID')])
        self.xsd_element(ao_node, 'Name', 'udt:TextType',
                         value=archive_object.name.value,
                         annotation=archive_object.name.user_annotation,
                         xsd_attributes=[XSDAttr('languageID', 'xsd:language')])
        if archive_object.appraisal_rule:
            self.xsd_appraisal_rule(ao_node, archive_object.appraisal_rule)
        if archive_object.access_restriction_code:
            self.xsd_access_restriction(ao_node, archive_object.access_restriction_code)
        if archive_object.content_description:
            self.xsd_content_description(ao_node, archive_object.content_description)
        self.xsd_children(ao_node, archive_object)
        return ao_node

    def xsd_document(self, parent, document):
        """Append XSD elements for the document to the given parent node."""
        document_node = self.xsd_element(parent, 'Document',
                                         cardinality=document.user_cardinality,
                                         annotation=document.user_annotation,
                                         xsd_attributes=[XSDAttr('Id', 'xsd:ID')])

        def safe_cardinality(entity):
            return getattr(entity, 'user_cardinality', None)

        def safe_concept_value(entity, rtype):
            return _concept_value(entity, rtype) if entity is not None else None

        self.xsd_element(document_node, 'Attachment', 'qdt:ArchivesBinaryObjectType',
                         xsd_attributes=[
                             XSDAttr('format', 'clmDAFFileTypeCode:FileTypeCodeType',
                                     cardinality=safe_cardinality(document.file_type_code),
                                     value=safe_concept_value(document.file_type_code,
                                                              'seda_file_type_code_value')),
                             XSDAttr('characterSetCode',
                                     'clmIANACharacterSetCode:CharacterSetCodeContentType',
                                     cardinality=safe_cardinality(document.character_set_code),
                                     value=safe_concept_value(document.character_set_code,
                                                              'seda_character_set_code_value')),
                             # hard-coded attributes
                             XSDAttr('mimeCode', 'clmIANAMIMEMediaType:MIMEMediaTypeContentType'),
                             XSDAttr('encodingCode',
                                     'clm60133:CharacterSetEncodingCodeContentType'),
                             XSDAttr('uri', 'xsd:anyURI'),
                             XSDAttr('filename', 'xsd:string'),
                         ])
        self.xsd_element(document_node, 'Type', 'qdt:CodeDocumentType',
                         value=_concept_value(document.document_type_code,
                                              'seda_document_type_code_value'),
                         xsd_attributes=[LIST_VERSION_ID_2009])
        if document.description:
            self.xsd_description(document_node, document.description)

    def xsd_appraisal_rule(self, parent, appraisal_rule):
        ar_node = self.xsd_element(parent, 'Appraisal',
                                   cardinality=appraisal_rule.user_cardinality,
                                   annotation=appraisal_rule.user_annotation,
                                   xsd_attributes=[XSDAttr('Id', 'xsd:ID')])
        ar_code = appraisal_rule.appraisal_rule_code
        ar_code_value = _concept_value(ar_code, 'seda_appraisal_rule_code_value')
        self.xsd_element(ar_node, 'Code', 'qdt:CodeAppraisalType',
                         value=ar_code_value,
                         annotation=ar_code.user_annotation,
                         xsd_attributes=[LIST_VERSION_ID_2009])
        ar_duration = appraisal_rule.seda_appraisal_rule_duration[0]
        ar_duration_value = _concept_value(ar_duration, 'seda_appraisal_rule_duration_value')
        self.xsd_element(ar_node, 'Duration', 'qdt:ArchivesDurationType',
                         value=ar_duration_value,
                         annotation=ar_duration.user_annotation)
        # hard-coded start date
        self.xsd_element(ar_node, 'StartDate', 'udt:DateType')

    access_restriction_tag_name = 'AccessRestrictionRule'

    def xsd_access_restriction(self, parent, access_restriction):
        """Append XSD elements for an access restriction to the given parent node."""
        ar_node = self.xsd_element(parent, self.access_restriction_tag_name,
                                   cardinality=access_restriction.user_cardinality,
                                   annotation=access_restriction.user_annotation,
                                   xsd_attributes=[XSDAttr('Id', 'xsd:ID')])
        ar_code = _concept_value(access_restriction, 'seda_access_restriction_code_value')
        self.xsd_element(ar_node, 'Code', 'qdt:CodeAccessRestrictionType',
                         value=ar_code,
                         xsd_attributes=[LIST_VERSION_ID_2009])
        # hard-coded start date
        self.xsd_element(ar_node, 'StartDate', 'udt:DateType')

    def xsd_content_description(self, parent, content_description):
        """Append XSD elements for a description content to the given parent node"""
        cd_node = self.xsd_element(parent, 'ContentDescription',
                                   cardinality=content_description.user_cardinality,
                                   annotation=content_description.user_annotation,
                                   xsd_attributes=[XSDAttr('Id', 'xsd:ID')])
        self.xsd_description_level(cd_node, content_description.description_level)
        for dt in ('oldest', 'latest'):
            date_entities = getattr(content_description, 'seda_%s_date' % dt)
            if date_entities:
                date_entity = date_entities[0]
                self.xsd_element(cd_node, '%sDate' % dt.capitalize(), 'udt:DateType',
                                 # default python formating '%Y-%m-%d' is fine
                                 value=date_entity.value,
                                 cardinality=date_entity.user_cardinality,
                                 annotation=date_entity.user_annotation)
        if content_description.description:
            self.xsd_description(cd_node, content_description.description)
        for keyword in content_description.keywords:
            self.xsd_keyword(cd_node, keyword)

    def xsd_description_level(self, parent, description_level):
        """Append XSD elements for a description level to the given parent node"""
        level = _concept_value(description_level, 'seda_description_level_value')
        self.xsd_element(parent, 'DescriptionLevel', 'qdt:CodeDescriptionLevelType',
                         annotation=description_level.user_annotation,
                         value=level,
                         xsd_attributes=[LIST_VERSION_ID_2009])

    def xsd_description(self, parent, description):
        """Append XSD elements for a description to the given parent node"""
        self.xsd_element(parent, 'Description', 'udt:TextType',
                         cardinality=description.user_cardinality,
                         annotation=description.user_annotation,
                         value=description.value,
                         xsd_attributes=[XSDAttr('languageID', 'xsd:language')])

    # extracted from xsd_keyword to allow parametrization for SEDA 1.0 vs 0.2 generation
    kw_tag_name = 'Keyword'
    kw_content_tag_type = 'qdt:KeywordContentType'
    kw_content_tag_attributes = [XSDAttr('role', 'xsd:token'),
                                 XSDAttr('languageID', 'xsd:language')]

    def xsd_keyword(self, parent, keyword):
        """Append XSD elements for a keyword to the given parent node"""
        kw_node = self.xsd_element(parent, self.kw_tag_name,
                                   cardinality=keyword.user_cardinality,
                                   annotation=keyword.user_annotation,
                                   xsd_attributes=[XSDAttr('Id', 'xsd:ID')])
        if keyword.concept:
            url = cwuri_url(keyword.concept)
            content = keyword.concept.label()
        else:
            url = content = None
        self.xsd_element(kw_node, 'KeywordContent', self.kw_content_tag_type,
                         value=content,
                         xsd_attributes=self.kw_content_tag_attributes)
        self.xsd_element(kw_node, 'KeywordReference', 'qdt:ArchivesIDType',
                         value=url,
                         xsd_attributes=self.xsd_attributes_scheme(keyword.scheme))

    @staticmethod
    def xsd_attributes_scheme(scheme=None):
        """Return a list of :class:`XSDAttr` for a scheme definition, with some proper values
        specified if a scheme is given.
        """
        attributes = [
            XSDAttr('schemeID', 'xsd:token'),
            XSDAttr('schemeName', 'xsd:string'),
            XSDAttr('schemeAgencyName', 'xsd:string'),
            XSDAttr('schemeVersionID', 'xsd:token'),
            XSDAttr('schemeDataURI', 'xsd:anyURI'),
            XSDAttr('schemeURI', 'xsd:anyURI'),
        ]
        if scheme is not None:
            # schemeURI
            attributes[-1]['fixed'] = cwuri_url(scheme)
            attributes[-1]['use'] = 'required'
            # schemeID
            attributes[0]['fixed'] = scheme.ark
            attributes[0]['use'] = 'required'
            # schemeName
            if scheme.title:
                attributes[1]['fixed'] = scheme.title
                attributes[1]['use'] = 'required'
        return attributes

    # generic methods ##############################################################################

    def xsd_children(self, parent, entity):
        """Iter on archive/archive object (unité documentaire) children, which may be either
        archive objects or documents, and append XSD elements for them to the given parent node.
        """
        for archive_object_or_document in entity.reverse_seda_parent:
            if archive_object_or_document.cw_etype == 'ProfileDocument':
                self.xsd_document(parent, archive_object_or_document)
            else:
                assert archive_object_or_document.cw_etype == 'ProfileArchiveObject'
                self.xsd_archive_object(parent, archive_object_or_document)

    def xsd_element(self, parent, name, xs_type=None, value=None, cardinality='1', annotation=None,
                    xsd_attributes=()):
        """Generic function to build a XSD element tag.

        Params:

        * `parent`, the parent etree node

        * `name`, value for the 'name' attribute of the xsd:element

        * `xs_type`, optional value for the 'type' attribute of the xsd:element

        * `value`, optional value for the 'fixed' attribute of the xsd:element

        * `cardinality`, optional string for the 'minOccurs' and/or 'maxOccurs' attributes of the
          xsd:element - may be one of '1', '0..1', '1..n', '0..n'

        * `annotation`, optional comment to add as annotation node for explanation of the element

        * `xsd_attributes`, list of :class:`XSDAttr` defining xsd:attributes children of the
          xsd:element
        """
        assert name, 'cannot build a XSD element tag with an empty name %r' % name
        attributes = {'name': name}
        if value is not None:
            attributes['fixed'] = unicode(value)
        if xs_type is not None and not xsd_attributes:
            attributes['type'] = xs_type
        assert cardinality in ('0..1', '0..n', '1', '1..n')
        if cardinality != '1':
            if cardinality[0] == '0':
                attributes['minOccurs'] = '0'
            if cardinality[-1] == 'n':
                attributes['maxOccurs'] = 'unbounded'
        element = self.element('xsd:element', parent, attributes)
        if annotation:
            annotation_node = self.element('xsd:annotation', element)
            self.element('xsd:documentation', annotation_node).text = annotation
        children_parent = None
        if xs_type is None:
            attributes_parent = self.element('xsd:complexType', element)
            children_parent = self.element('xsd:sequence', attributes_parent)
        elif xsd_attributes:
            ct = self.element('xsd:complexType', element)
            scontent = self.element('xsd:simpleContent', ct)
            attributes_parent = self.element('xsd:extension', scontent, {'base': xs_type})
        for xsd_attr in xsd_attributes:
            self.element('xsd:attribute', attributes_parent, xsd_attr)
        return children_parent

    # compatibility testing ########################################################################

    def blockers(self):
        """Yield messages describing a problem that prevents the given profile to be exported."""
        profile = self.entity
        if not profile.archives:
            yield (profile.eid, _('the profile should have at least a document unit'))
        else:
            for archive in profile.archives:
                if archive.user_cardinality in ('0..1', '0..n'):
                    yield (archive.eid, _('0..1 and 0..n cardinalities are forbidden on '
                                          'first-level document unit'))
                if not archive.content_description:
                    yield (archive.eid,
                           _('first level document unit must have a content description defined'))
                elif archive.content_description.user_cardinality == '0..1':
                    yield (archive.eid, _('0..1 cardinality is not allowed on content '
                                          'description of a first-level document unit'))
                if not archive.access_restriction_code:
                    yield (archive.eid, _('first level document unit must have an access '
                                          'restriction defined'))
                elif archive.access_restriction_code.user_cardinality == '0..1':
                    yield (archive.eid, _('0..1 cardinality is not allowed on access '
                                          'restriction of a first-level document unit'))

    def is_compatible(self):
        """Return True if the adapted entity may be exported in the format handled by this adapter.
        """
        try:
            next(self.blockers())
            return False
        except StopIteration:
            return True


def _concept_value(entity, relation_name):
    """Return the value of the concept linked to the entity through the given relation or None
    """
    rset = entity.related(relation_name)
    if rset:
        return rset.one().label()
    return None


class SEDA02XSDExport(SEDA1XSDExport):
    """Adapter to build an XSD representation of a SEDA profile, using SEDA 0.2 specification"""
    __regid__ = 'SEDA-0.2.xsd'

    namespaces = SEDA1XSDExport.namespaces.copy()
    namespaces[None] = 'fr:gouv:ae:archive:draft:standard_echange_v0.2'
    namespaces['qdt'] = 'fr:gouv:ae:archive:draft:standard_echange_v0.2:QualifiedDataType:1'
    namespaces['udt'] = 'urn:un:unece:uncefact:data:standard:UnqualifiedDataType:6'

    root_attributes = SEDA1XSDExport.root_attributes.copy()
    root_attributes['targetNamespace'] = 'fr:gouv:ae:archive:draft:standard_echange_v0.2'
    root_attributes['version'] = '1.1'

    def xsd_archive(self, parent, archive):
        """Append XSD elements for an archive to the given parent node."""
        archive_node = self.xsd_element(parent, 'Contains',
                                        cardinality=archive.user_cardinality,
                                        annotation=archive.user_annotation,
                                        xsd_attributes=[XSDAttr('Id', 'xsd:ID')])
        # hard-coded description's language
        self.xsd_element(archive_node, 'DescriptionLanguage', 'qdt:CodeLanguageType',
                         value='fr',
                         xsd_attributes=[LIST_VERSION_ID_2009])
        self.xsd_element(archive_node, 'Name', 'udt:TextType',
                         value=archive.seda_name[0].value,
                         annotation=archive.seda_name[0].user_annotation,
                         xsd_attributes=[XSDAttr('languageID', 'xsd:language')])
        # in SEDA 0.2, description level is on the archive element, not on its content description
        content_description = archive.content_description
        self.xsd_description_level(archive_node, content_description.description_level)
        if archive.appraisal_rule:
            self.xsd_appraisal_rule(archive_node, archive.appraisal_rule)
        # in SEDA 0.2, access restriction is not mandatory, though it is in our model
        self.xsd_access_restriction(archive_node, archive.access_restriction_code)
        self.xsd_content_description(archive_node, content_description)
        self.xsd_children(archive_node, archive)

    def xsd_content_description(self, parent, content_description):
        """Append XSD elements for a description content to the given parent node"""
        cd_node = self.xsd_element(parent, 'ContentDescription',
                                   cardinality=content_description.user_cardinality,
                                   annotation=content_description.user_annotation,
                                   xsd_attributes=[XSDAttr('Id', 'xsd:ID')])
        for dt in ('oldest', 'latest'):
            date_entities = getattr(content_description, 'seda_%s_date' % dt)
            if date_entities:
                date_entity = date_entities[0]
                self.xsd_element(cd_node, '%sDate' % dt.capitalize(), 'udt:DateType',
                                 # default python formating '%Y-%m-%d' is fine
                                 value=date_entity.value,
                                 cardinality=date_entity.user_cardinality,
                                 annotation=date_entity.user_annotation)
        if content_description.description:
            self.xsd_description(cd_node, content_description.description)
        # in SEDA 0.2, content description has a description language as well
        self.xsd_element(cd_node, 'DescriptionLanguage', 'qdt:CodeLanguageType',
                         value='fr',
                         xsd_attributes=[LIST_VERSION_ID_2009])
        for keyword in content_description.keywords:
            self.xsd_keyword(cd_node, keyword)
        # in SEDA 0.2, there may be some access restriction on the content description but it's on
        # the archive in our model

    def xsd_archive_object(self, parent, archive_object):
        """Append XSD elements for the archive object to the given parent node."""
        ao_node = super(SEDA02XSDExport, self).xsd_archive_object(parent, archive_object)
        # in SEDA 0.2, description level is on the archive object element, not on its content
        # description
        if archive_object.content_description:
            content_description = archive_object.content_description
            self.xsd_description_level(ao_node, content_description.description_level)
        else:
            self.xsd_element(ao_node, 'DescriptionLevel', 'qdt:CodeDescriptionLevelType',
                             xsd_attributes=[LIST_VERSION_ID_2009])

    # in SEDA 0.2, ArchiveObject tag name is 'Contains' (as for Archive)
    archive_object_tag_name = 'Contains'
    # in SEDA 0.2, AccessRestrictionRule tag name is 'AccessRestriction'
    access_restriction_tag_name = 'AccessRestriction'
    # in SEDA 0.2, keyword tag name is 'ContentDescriptive', not 'Keyword' and keyword content type
    # is TextType and there is no 'role' attribute
    kw_tag_name = 'ContentDescriptive'
    kw_content_tag_type = 'udt:TextType'
    kw_content_tag_attributes = [XSDAttr('languageID', 'xsd:language')]



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref entity's classes"""

from lxml import etree

from logilab.common.decorators import monkeypatch

from cubicweb.view import Adapter, EntityAdapter
from cubicweb.predicates import match_kwargs, relation_possible, is_instance
from cubicweb.entities import AnyEntity, fetch_config, authobjs


class ARKGeneratorMixIn(object):
    """Entity adapter for ARK unique identifier generation"""
    __abstract__ = True
    __regid__ = 'IARKGenerator'
    __select__ = match_kwargs('naa_what')

    def generate_ark(self):
        """Return a new ARK identifier as unicode"""
        return u'{0}/{1}'.format(self.cw_extra_kwargs['naa_what'], self.assign_name())

    def assign_name(self):
        """Assign and return a new name part of the ARK identifier"""
        raise NotImplementedError()


class ARKExtIdentifierGenerator(ARKGeneratorMixIn, Adapter):
    """Simple adapter for allocation of ark to object which are not (yet) entities."""

    def assign_name(self):
        """Return a new unique identifier as unicode"""
        dbh = self._cw.repo.system_source.dbhelper
        for sql in dbh.sqls_increment_sequence('ext_ark_count'):
            cu = self._cw.system_sql(sql)
        count = cu.fetchone()[0]
        return u'a{0:09d}'.format(count)


class ARKCWIdentifierGenerator(ARKGeneratorMixIn, Adapter):
    """saem_ref.IARKGenerator entity adapter generating ARK like identifier during non-production
    phase.
    """
    __select__ = ARKGeneratorMixIn.__select__ & match_kwargs('eid', 'etype')

    def assign_name(self):
        eid = self.cw_extra_kwargs['eid']
        etype = self.cw_extra_kwargs['etype']
        prefix = {
            'AuthorityRecord': u'r',
            'SEDAProfile': u'p',
            'Concept': u'c',
            'ConceptScheme': u'v',
            'Organization': u'o',
            'Agent': u'oa',
            'OrganizationUnit': u'ou',
        }.get(etype, u'')
        return u'{0}{1:09d}'.format(prefix, eid)


class ArkNAALocator(EntityAdapter):
    """Adapter responsible to retrieve the proper ARK Name Assigning Authority depending on the
    entity type
    """
    __abstract__ = True
    __regid__ = 'IArkNAALocator'

    def naa_what(self):
        """Return the ARK NameAssigningAuthority entity or None if not specified"""
        raise NotImplementedError()


class DirectArkNAALocator(ArkNAALocator):
    """Return NAA specified through the ark_naa relation"""
    __select__ = relation_possible('ark_naa')

    def naa_what(self):
        # entity is usually not yet created, since ark has to be generated before entity creation
        if 'ark_naa' in getattr(self.entity, 'cw_edited', {}):
            return self._cw.entity_from_eid(self.entity.cw_edited['ark_naa']).what
        elif self.entity.ark_naa:
            return self.entity.ark_naa[0].what
        return None


class AgentArkNAALocator(ArkNAALocator):
    """Return NAA specified through the authority to which the agent belong"""
    __select__ = is_instance('Agent', 'OrganizationUnit')

    def naa_what(self):
        # entity is usually not yet created, since ark has to be generated before entity creation
        if 'authority' in getattr(self.entity, 'cw_edited', {}):
            authority = self._cw.entity_from_eid(self.entity.cw_edited['authority'])
            if authority.ark_naa:
                return authority.ark_naa[0].what
        elif self.entity.authority and self.authority[0].ark_naa:
            return self.authority[0].ark_naa[0].what
        return None


class ConceptArkNAALocator(ArkNAALocator):
    """Return NAA for Concept, usually specified through a record in transaction data"""
    __select__ = is_instance('Concept')

    def naa_what(self):
        if 'concept_scheme' in self._cw.transaction_data:
            scheme = self._cw.entity_from_eid(self._cw.transaction_data['concept_scheme'])
            return scheme.cw_adapt_to(self.__regid__).naa_what()
        elif self.entity.in_scheme:  # entity may not yet be related to its scheme
            return self.entity.in_scheme[0].cw_adapt_to(self.__regid__).naa_what()
        return None


class ExternalUri(AnyEntity):
    __regid__ = 'ExternalUri'
    fetch_attrs, cw_fetch_order = fetch_config(('uri',))


@monkeypatch(authobjs.CWUser, methodname='naa')
@property
def naa(self):
    if self.authority and self.authority[0].ark_naa:
        return self.authority[0].ark_naa[0]
    return None


def substitute_xml_prefix(prefix_name, namespaces):
    """Given an XML prefixed name in the form `'ns:name'`, return the string `'{<ns_uri>}name'`
    where `<ns_uri>` is the URI for the namespace prefix found in `namespaces`.

    This new string is then suitable to build an LXML etree.Element object.

    Example::

        >>> substitude_xml_prefix('xlink:href', {'xlink': 'http://wwww.w3.org/1999/xlink'})
        '{http://www.w3.org/1999/xlink}href'

    """
    try:
        prefix, name = prefix_name.split(':', 1)
    except ValueError:
        return prefix_name
    assert prefix in namespaces, 'Unknown namespace prefix: {0}'.format(prefix)
    return '{{{0}}}'.format(namespaces[prefix]) + name


class AbstractXmlAdapter(EntityAdapter):
    """Abstract adapter to produce XML documents."""

    content_type = 'text/xml'
    encoding = 'utf-8'
    namespaces = {}

    @property
    def file_name(self):
        """Return a file name for the dump."""
        raise NotImplementedError

    def dump(self):
        """Return an XML string for the adapted entity."""
        raise NotImplementedError

    def element(self, tag, parent=None, attributes=None, text=None):
        """Generic function to build a XSD element tag.

        Params:

        * `name`, value for the 'name' attribute of the xsd:element

        * `parent`, the parent etree node

        * `attributes`, dictionary of attributes
        """
        attributes = attributes or {}
        tag = substitute_xml_prefix(tag, self.namespaces)
        for attr, value in attributes.items():
            newattr = substitute_xml_prefix(attr, self.namespaces)
            attributes[newattr] = value
            if newattr != attr:
                attributes.pop(attr)
        if parent is None:
            elt = etree.Element(tag, attributes, nsmap=self.namespaces)
        else:
            elt = etree.SubElement(parent, tag, attributes)
        if text is not None:
            elt.text = text
        return elt



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref monkeypatch of not-yet-integrated patches"""


# https://www.cubicweb.org/ticket/12512052 #########################################################
# this is not probably the right fix, see also https://www.cubicweb.org/ticket/12512176

from cubicweb.web import NoSelectableObject  # noqa
from cubicweb.web.views import ajaxcontroller  # noqa


@ajaxcontroller.ajaxfunc(output_type='xhtml')
def render(self, registry, oid, eid=None, selectargs=None, renderargs=None):
    if eid is not None:
        rset = self._cw.eid_rset(eid)
        # XXX set row=0
    elif self._cw.form.get('rql'):
        rset = self._cw.execute(self._cw.form['rql'])
    else:
        rset = None
    try:
        viewobj = self._cw.vreg[registry].select(oid, self._cw, rset=rset,
                                                 **ajaxcontroller.optional_kwargs(selectargs))
    except NoSelectableObject:
        return u''
    return self._call_view(viewobj, **ajaxcontroller.optional_kwargs(renderargs))


def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__, (render,))
    vreg.register_and_replace(render, ajaxcontroller.render)



# coding: utf-8
# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""Tests for AuthorityRecord entities"""

import datetime

from mock import MagicMock, patch
from lxml import etree

from cubicweb import Binary
from cubicweb.devtools.testlib import CubicWebTC

import testutils


class EACExportTC(CubicWebTC):
    """Unit tests for EAC-CPF exports."""

    def test_richstring_plain(self):
        with self.admin_access.cnx() as cnx:
            alice = testutils.authority_record(cnx, u'Alice')
            desc = u'ding'
            mandate = cnx.create_entity(
                'Mandate', term=u'ding-girl',
                description=desc, description_format=u'text/plain',
                mandate_agent=alice)
            cnx.commit()
            serializer = alice.cw_adapt_to('EAC-CPF')
            tag, = serializer._eac_richstring_paragraph_elements(
                mandate, 'description')
        self.assertEqual(tag.tag, 'p')
        self.assertEqual(tag.text, desc)

    def test_richstring_html_simple(self):
        with self.admin_access.cnx() as cnx:
            alice = testutils.authority_record(cnx, u'Alice')
            desc = u'<span>ding</span>'
            mandate = cnx.create_entity(
                'Mandate', term=u'ding-girl',
                description=desc, description_format=u'text/html',
                mandate_agent=alice)
            cnx.commit()
            serializer = alice.cw_adapt_to('EAC-CPF')
            tag, = serializer._eac_richstring_paragraph_elements(
                mandate, 'description')
        self.assertEqual(tag.tag, 'span')
        self.assertIn(desc, etree.tostring(tag))

    def test_richstring_html_multiple_elements(self):
        with self.admin_access.cnx() as cnx:
            alice = testutils.authority_record(cnx, u'Alice')
            desc = [u'<h1>she <i>rules!</i></h1>', u'<a href="1">pif</a>']
            mandate = cnx.create_entity(
                'Mandate', term=u'chairgirl',
                description=u''.join(desc), description_format=u'text/html',
                mandate_agent=alice)
            cnx.commit()
            serializer = alice.cw_adapt_to('EAC-CPF')
            h1, a = serializer._eac_richstring_paragraph_elements(
                mandate, 'description')
        self.assertEqual(h1.tag, 'h1')
        self.assertEqual(a.tag, 'a')
        self.assertIn(etree.tostring(h1), desc[0])
        self.assertIn(etree.tostring(a), desc[1])

    def test_richstring_markdown(self):
        with self.admin_access.cnx() as cnx:
            alice = testutils.authority_record(cnx, u'Alice')
            desc = u'[pif](http://gadget.com) is *red*'
            desc_html = (
                u'<a href="http://gadget.com">pif</a> '
                u'is <em>red</em>'
            )
            mandate = cnx.create_entity(
                'Mandate', term=u'w',
                description=desc, description_format=u'text/markdown',
                mandate_agent=alice)
            cnx.commit()
            serializer = alice.cw_adapt_to('EAC-CPF')
            tag, = serializer._eac_richstring_paragraph_elements(
                mandate, 'description')
        self.assertEqual(tag.tag, 'p')
        self.assertIn(desc_html, etree.tostring(tag))

    def test_richstring_rest(self):
        with self.admin_access.cnx() as cnx:
            alice = testutils.authority_record(cnx, u'Alice')
            desc = u'`pif <http://gadget.com>`_ is *red*'
            desc_html = (
                u'<a class="reference" href="http://gadget.com">pif</a> '
                u'is <em>red</em>'
            )
            mandate = cnx.create_entity(
                'Mandate', term=u'w',
                description=desc, description_format=u'text/rest',
                mandate_agent=alice)
            cnx.commit()
            serializer = alice.cw_adapt_to('EAC-CPF')
            ptag, = serializer._eac_richstring_paragraph_elements(
                mandate, 'description')
        self.assertEqual(ptag.tag, 'p')
        self.assertIn(desc_html, etree.tostring(ptag))

    def test_richstring_empty(self):
        def check(authority_record):
            serializer = authority_record.cw_adapt_to('EAC-CPF')
            res = serializer._eac_richstring_paragraph_elements(
                mandate, 'description')
            self.assertEqual(res, [])

        with self.admin_access.cnx() as cnx:
            alice = testutils.authority_record(cnx, u'Alice')
            mandate = cnx.create_entity(
                'Mandate', term=u'w',
                description=None,
                mandate_agent=alice)
            cnx.commit()
            check(alice)
        with self.admin_access.cnx() as cnx:
            cnx.execute(
                'SET X description_format "text/rest" WHERE X is Mandate')
            cnx.commit()
            authority_record = cnx.find('AuthorityRecord', name=u'Alice').one()
            check(authority_record)


class AuthorityRecordExportTC(CubicWebTC, testutils.XmlTestMixin):
    """Functional test case for AuthorityRecord export in EAC-CPF."""

    @patch('cubes.saem_ref.hooks.datetime')
    def setup_database(self, mock_dt):
        mock_dt.utcnow.return_value = datetime.datetime(2016, 2, 1)
        with self.admin_access.client_cnx() as cnx:
            work_address2 = cnx.create_entity('PostalAddress', street=u"1 av. de l'europe",
                                              postalcode=u'31400', city=u'Toulouse')
            work_address3 = cnx.create_entity('PostalAddress', street=u'104 bd L.-A. Blanqui',
                                              postalcode=u'75013', city=u'Paris')
            home_address = cnx.create_entity('PostalAddress', street=u'Place du Capitole',
                                             postalcode=u'31000', city=u'Toulouse')
            agent1 = testutils.authority_record(cnx, u'Alice')
            agent2 = testutils.authority_record(cnx, u'Super Service',
                                                kind=u'authority')
            cnx.create_entity('AgentPlace', role=u'work', place_agent=agent2,
                              place_address=work_address2)
            agent3 = testutils.authority_record(cnx, u'Charlie', ark=u'1234')
            cnx.create_entity('AgentPlace', role=u'work', place_agent=agent3,
                              place_address=work_address3)
            cnx.create_entity('AgentPlace', role=u'home', place_agent=agent3,
                              place_address=home_address)
            cnx.create_entity('ChronologicalRelation',
                              chronological_predecessor=agent2,
                              xml_wrap=Binary('<plip>plop</plip>'),
                              chronological_successor=agent3)
            logilab_uri = cnx.create_entity('ExternalUri', uri=u'http://www.logilab.fr')
            cnx.create_entity('EACResourceRelation',
                              resource_relation_resource=logilab_uri,
                              resource_relation_agent=agent3,
                              xml_wrap=Binary('<pif><paf>pouf</paf></pif>'))
            gironde = testutils.authority_record(
                cnx, u'Gironde. Conseil général', kind=u'authority',
                ark=u'FRAD033_EAC_00001',
                start_date=datetime.date(1800, 1, 1),
                end_date=datetime.date(2099, 1, 1),
                isni=u'22330001300016')
            delphine = testutils.agent(cnx, u'Delphine Jamet')
            adm_dir = testutils.authority_record(
                cnx, (u"Gironde. Conseil général. Direction de "
                      u"l'administration et de la sécurité juridique"),
                kind=u'unknown-agent-kind')
            cg32 = testutils.authority_record(cnx, u'CG32', kind=u'unknown-agent-kind')
            trash = testutils.authority_record(cnx, u'Trash', kind=u'unknown-agent-kind')
            place_uri = cnx.create_entity('ExternalUri',
                                          uri=u'http://catalogue.bnf.fr/ark:/12148/cb152418385',
                                          cwuri=u'http://catalogue.bnf.fr/ark:/12148/cb152418385')
            agent_x = cnx.create_entity('ExternalUri', uri=u'agent-x', cwuri=u'agent-x')
            social_action_uri = cnx.create_entity(
                'ExternalUri', uri=u'http://data.culture.fr/thesaurus/page/ark:/67717/T1-200',
                cwuri=u'http://data.culture.fr/thesaurus/page/ark:/67717/T1-200')
            environment_uri = cnx.create_entity(
                'ExternalUri', uri=u'http://data.culture.fr/thesaurus/page/ark:/67717/T1-1074',
                cwuri=u'http://data.culture.fr/thesaurus/page/ark:/67717/T1-1074')
            resource_uri = cnx.create_entity(
                'ExternalUri', uri=u'http://gael.gironde.fr/ead.html?id=FRAD033_IR_N',
                cwuri=u'http://gael.gironde.fr/ead.html?id=FRAD033_IR_N')
            meeting_uri = cnx.create_entity(
                'ExternalUri', uri=u'http://example.org/meeting',
                cwuri=u'http://example.org/meeting')
            cnx.create_entity('EACSource', title=u'1. Ouvrages imprimés...',
                              description=u'des bouquins', source_agent=gironde)
            cnx.create_entity('EACSource', title=u'Site des Archives départementales de la Gironde',
                              url=u'http://archives.gironde.fr', source_agent=gironde,
                              xml_wrap=Binary('<some>thing</some>'))
            cnx.create_entity('Activity', type=u'create',
                              start=datetime.datetime(2013, 4, 24, 5, 34, 41),
                              end=datetime.datetime(2013, 4, 24, 5, 34, 41),
                              description=u'bla bla bla', generated=gironde)
            cnx.create_entity('Activity', type=u'modify', associated_with=delphine,
                              start=datetime.datetime(2015, 1, 15, 7, 16, 33),
                              end=datetime.datetime(2015, 1, 15, 7, 16, 33),
                              generated=gironde)
            gironde_address = cnx.create_entity('PostalAddress',
                                                street=u'1 Esplanade Charles de Gaulle',
                                                postalcode=u'33074', city=u'Bordeaux Cedex')
            cnx.create_entity('AgentPlace', role=u'siege', name=u'Bordeaux (Gironde, France)',
                              place_agent=gironde, place_address=gironde_address,
                              equivalent_concept=place_uri)
            cnx.create_entity('AgentPlace', role=u'domicile', name=u'Toulouse (France)',
                              place_agent=gironde)
            cnx.create_entity('AgentPlace', role=u'dodo', name=u'Lit',
                              place_agent=gironde)
            cnx.create_entity('LegalStatus', term=u'Collectivité territoriale',
                              start_date=datetime.date(1234, 1, 1),
                              end_date=datetime.date(3000, 1, 1),
                              description=u'Description du statut',
                              legal_status_agent=gironde,
                              has_citation=[cnx.create_entity('Citation', note=u"legal foo")])
            cnx.create_entity('Mandate', term=u'1. Constitutions françaises',
                              description=u'Description du mandat',
                              mandate_agent=gironde)
            cnx.create_entity(
                'History',
                text=(
                    u"<p>La loi du 22 décembre 1789, en divisant ...</p>\n"
                    u"<p>L'inspecteur Canardo</p>"
                ),
                text_format=u"text/html",
                history_agent=gironde,
                has_citation=[
                    cnx.create_entity('Citation', uri=(
                        u"http://www.assemblee-nationale.fr/histoire/images-decentralisation/"
                        u"decentralisation/loi-du-22-decembre-1789-.pdf"
                    )),
                    cnx.create_entity('Citation',
                                      uri=u"http://pifgadget", note=u"Voir aussi pifgadget"),
                ],
            )
            cnx.create_entity('GeneralContext',
                              content=u'<p>sous une pluie battante</p>',
                              content_format=u'text/html',
                              has_citation=cnx.create_entity(
                                  'Citation', uri=u'http://meteoplouf.net'),
                              general_context_of=gironde)
            cnx.create_entity('Structure',
                              description=u'Pour accomplir ses missions ...',
                              structure_agent=gironde)
            cnx.create_entity('AgentFunction',
                              description=u'Quatre grands domaines de compétence...',
                              function_agent=gironde)
            cnx.create_entity('AgentFunction', name=u'action sociale',
                              description=u'1. Solidarité\nblablabla.',
                              function_agent=gironde, equivalent_concept=social_action_uri)
            cnx.create_entity('AgentFunction', name=u'environnement',
                              function_agent=gironde, equivalent_concept=environment_uri)
            cnx.create_entity('Occupation', term=u'Réunioniste',
                              description=u'Organisation des réunions ...',
                              start_date=datetime.date(1987, 1, 1),
                              end_date=datetime.date(2099, 1, 1),
                              occupation_agent=gironde, equivalent_concept=meeting_uri)
            cnx.create_entity('HierarchicalRelation', hierarchical_parent=adm_dir,
                              hierarchical_child=gironde, description=u'Coucou',
                              start_date=datetime.date(2008, 1, 1),
                              end_date=datetime.date(2099, 1, 1))
            cnx.create_entity('ChronologicalRelation', chronological_predecessor=cg32,
                              chronological_successor=gironde)
            cnx.create_entity('ChronologicalRelation', chronological_predecessor=gironde,
                              chronological_successor=trash)
            cnx.create_entity('AssociationRelation', association_from=gironde,
                              association_to=agent_x)
            cnx.create_entity('EACResourceRelation', agent_role=u'creatorOf',
                              resource_role=u"Fonds d'archives",
                              start_date=datetime.date(1673, 1, 1),
                              end_date=datetime.date(1963, 1, 1),
                              resource_relation_resource=resource_uri,
                              resource_relation_agent=gironde)
            cnx.commit()
            gironde.cw_adapt_to('IWorkflowable').fire_transition('publish')
            cnx.commit()
            self.agent1_eid = agent1.eid
            self.agent2_eid = agent2.eid
            self.agent3_eid = agent3.eid
            self.gironde_eid = gironde.eid
            self.adm_dir_eid = adm_dir.eid
            self.cg32_eid = cg32.eid
            self.trash_eid = trash.eid
            self.agent_x_eid = agent_x.eid
            self.work_address2_eid = work_address2.eid
            self.work_address3_eid = work_address3.eid
            self.home_address_eid = home_address.eid

    def test_agent_eac_export_simple(self):
        # Given an agent (created in setup_database), export it to EAC-CPF
        with self.admin_access.client_cnx() as cnx:
            agent3 = cnx.entity_from_eid(self.agent3_eid)
            agent2 = cnx.entity_from_eid(self.agent2_eid)
            agent2.absolute_url = MagicMock(return_value=u'http://www.example.org/agent2')
            agent3_eac = agent3.cw_adapt_to('EAC-CPF').dump()
        # Then check that output XML is as expected
        with open(self.datapath('EAC/agent3_export.xml')) as f:
            expected_eac = f.read().replace("@@@EID_GOES_HERE@@@", str(self.agent3_eid))
        self.assertXmlEqual(agent3_eac, expected_eac)
        self.assertXmlValid(agent3_eac, self.datapath('cpf.xsd'))

    def test_agent_eac_export_full(self):
        with self.admin_access.client_cnx() as cnx:
            for eid, absolute_url in (
                (self.adm_dir_eid, u'http://www.example.org/adm_dir'),
                (self.cg32_eid, u'http://www.example.org/cg32'),
                (self.trash_eid, u"/dev/null"),
                (self.agent_x_eid, u'http://www.example.org/agent_x'),
            ):
                cnx.entity_from_eid(eid).absolute_url = MagicMock(return_value=absolute_url)
            gironde = cnx.entity_from_eid(self.gironde_eid)
            gironde_eac = gironde.cw_adapt_to('EAC-CPF').dump()
        with open(self.datapath('EAC/FRAD033_EAC_00001_simplified_export.xml')) as f:
            expected_eac = f.read().replace("@@@EID_GOES_HERE@@@", str(self.gironde_eid))
        with open('/tmp/gironde.xml', 'w') as f:
            f.write(gironde_eac)
        self.assertXmlEqual(gironde_eac, expected_eac)
        self.assertXmlValid(gironde_eac, self.datapath('cpf.xsd'))


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem_ref unit tests for schema"""

import sqlite3
import unittest
from datetime import date
from contextlib import contextmanager

from cubicweb import ValidationError, Unauthorized, neg_role
from cubicweb.devtools.testlib import CubicWebTC

from cubes.saem_ref import (PERMISSIONS_GRAPHS, mandatory_rdefs,
                            optional_relations)

import testutils


def graph_relations(schema, parent_structure):
    """Given a parent structure of a composite graph (and a schema object),
    return relation information `(rtype, role)` sets where `role` is the role
    of the child in the relation for the following kinds of relations:

    * structural relations,
    * optional relations (cardinality of the child not in '1*'),
    * mandatory relations (cardinality of the child in '1*').
    """
    def concat_sets(sets):
        """Concatenate sets"""
        return reduce(lambda x, y: x | y, sets, set())

    optionals = concat_sets(
        optional_relations(schema, parent_structure).values())
    mandatories = set([
        (rdef.rtype, neg_role(role))
        for rdef, role in mandatory_rdefs(schema, parent_structure)])
    structurals = concat_sets(map(set, parent_structure.values()))
    return structurals, optionals, mandatories


@contextmanager
def assertValidationError(self, cnx):
    with self.assertRaises(ValidationError) as cm:
        yield cm
        cnx.commit()
    cnx.rollback()


@contextmanager
def assertUnauthorized(self, cnx):
    with self.assertRaises(Unauthorized) as cm:
        yield cm
        cnx.commit()
    cnx.rollback()


def assertUnauthorizedRQL(self, cnx, rql):
    with assertUnauthorized(self, cnx):
        cnx.execute(rql)


def create_published_agents_panel(cnx):
    for name, kind in (
        (u"Mr Pink", u"person"),
        (u"Adams", u"family"),
        (u"Direction de la communication", u"authority"),
    ):
        ar = testutils.authority_record(cnx, name, kind)
        cnx.commit()
        ar.cw_adapt_to('IWorkflowable').fire_transition('publish')
        cnx.commit()


class SchemaConstraintsTC(CubicWebTC):
    assertValidationError = assertValidationError

    def test_on_create_set_end_date_before_start_date(self):
        """ create an entity whose end_date is before start_date.
        ValidationError expected
        """
        with self.admin_access.repo_cnx() as cnx:
            with self.assertValidationError(cnx) as cm:
                testutils.authority_record(cnx, u'Arthur',
                                           start_date=date(524, 02, 9),
                                           end_date=date(500, 07, 12))
            self.assertIn("must be less than", str(cm.exception))

    def test_on_update_set_end_date_before_start_date(self):
        """ create a valid entity and update it with a new end_date set before the start_date.
            ValidationError expected
        """
        if sqlite3.sqlite_version_info < (3, 7, 12):
            # with sqlite earlier than 3.7.12, boundary constraints are not checked by the database,
            # hence the constraint is only triggered on start_date modification
            self.skipTest('unsupported sqlite version')
        with self.admin_access.repo_cnx() as cnx:
            agent = testutils.authority_record(cnx, u'Arthur',
                                               start_date=date(454, 02, 9),
                                               end_date=date(475, 04, 12))
            cnx.commit()
            with self.assertValidationError(cnx) as cm:
                agent.cw_set(end_date=date(442, 07, 12))
            self.assertIn("must be less than", str(cm.exception))

    def test_on_update_set_start_date_after_end_date(self):
        """ create an entity without start_date :
            No constraint on the end_date
            update the entity with a start_date set after the start_date :
            ValidationError expected
        """
        with self.admin_access.repo_cnx() as cnx:
            agent = testutils.authority_record(cnx, u'Arthur', end_date=date(476, 02, 9))
            cnx.commit()
            with self.assertValidationError(cnx) as cm:
                agent.cw_set(start_date=date(527, 04, 12))
            self.assertIn("must be less than", str(cm.exception))

    def test_published_constraint_on_contact_point(self):
        """ create two agents: one published P and one not published N.
            create one OU and check that interface will only show P that can become its
            contact point
        """
        with self.admin_access.repo_cnx() as cnx:
            peter = testutils.agent(cnx, u'Peter')
            testutils.agent(cnx, u'Norton')
            cnx.commit()
            peter.cw_adapt_to('IWorkflowable').fire_transition('publish')
            ou = testutils.organization_unit(cnx, u'Alice')
            cnx.commit()
            rset = ou.unrelated('contact_point', 'Agent')
            self.assertEqual(rset.one().eid, peter.eid)

    def test_published_constraint_on_archival_agent(self):
        """ create two OU: one published P and one not published N.
            create one Organization and check that interface will only show P that can become its
            archival agent
        """
        with self.admin_access.repo_cnx() as cnx:
            pou = testutils.organization_unit(cnx, u'P OU', archival_roles=['archival'])
            testutils.organization_unit(cnx, u'N OU', archival_roles=['archival'])
            cnx.commit()
            pou.cw_adapt_to('IWorkflowable').fire_transition('publish')
            cnx.commit()
            authority = cnx.find('Organization', name=u'Default authority').one()
            rset = authority.unrelated('archival_unit', 'OrganizationUnit')
            self.assertEqual(rset.one().eid, pou.eid)

    def test_agent_authority_record_is_a_person(self):
        """Authority record for an agent is limited to person"""
        with self.admin_access.repo_cnx() as cnx:
            create_published_agents_panel(cnx)
            peter = testutils.agent(cnx, u'Peter')
            self.assertEqual(
                peter.unrelated('authority_record', 'AuthorityRecord').one(),
                cnx.find("AuthorityRecord", name=u"Mr Pink").one(),
            )

    def test_organization_unit_authority_record_is_a_authority(self):
        """Authority record for an organization unit is limited to authority"""
        with self.admin_access.repo_cnx() as cnx:
            create_published_agents_panel(cnx)
            pou = testutils.organization_unit(cnx, u'P OU')
            self.assertEqual(
                pou.unrelated('authority_record', 'AuthorityRecord').one(),
                cnx.find("AuthorityRecord", name=u"Direction de la communication").one(),
            )

    def test_organization_unit_contact_point_in_the_same_authority(self):
        """Create two agents on two distinct authorities. Create an organization unit and check that
        interface will only show consistent proposal for contact point
        """
        with self.admin_access.repo_cnx() as cnx:
            # Create an archival agent in a custom authority
            authority = cnx.create_entity('Organization', name=u'boss club')
            jdoe = testutils.agent(cnx, u'jdoe', authority=authority)
            # Create another archival agent in the default authority
            norton = testutils.agent(cnx, u'Norton')
            cnx.commit()
            # Publish both agents (because of the published constraint)
            jdoe.cw_adapt_to('IWorkflowable').fire_transition('publish')
            norton.cw_adapt_to('IWorkflowable').fire_transition('publish')
            cnx.commit()
            # Now create an organisation unit and check the unrelated list
            ou = testutils.organization_unit(cnx, u'Alice')
            cnx.commit()
            rset = ou.unrelated('contact_point', 'Agent')
            self.assertEqual(rset.one().eid, norton.eid)

    def test_organization_archival_unit_in_whatever_authority(self):
        """Create two organizations on two distinct authorities. Check that interface will show both
        of them as proposal for archival_agent of the default organization
        """
        with self.admin_access.repo_cnx() as cnx:
            # Create an organization unit in a custom authority
            authority = cnx.create_entity('Organization', name=u'boss club')
            jdoe = testutils.organization_unit(cnx, u'jdoe', archival_roles=['archival'])
            # Create another archival organization unit in the default authority
            norton = testutils.organization_unit(cnx, u'Norton', archival_roles=[u'archival'])
            cnx.commit()
            # Publish both organization units (because of the published constraint)
            jdoe.cw_adapt_to('IWorkflowable').fire_transition('publish')
            norton.cw_adapt_to('IWorkflowable').fire_transition('publish')
            cnx.commit()
            # Now check the unrelated list for authority's archival_unit
            authority = cnx.find('Organization', name=u'Default authority').one()
            rset = authority.unrelated('archival_unit', 'OrganizationUnit')
            self.assertEqual(len(rset), 2)

    def test_agent_authority_consistency(self):
        with self.admin_access.repo_cnx() as cnx:
            authority1 = cnx.find('Organization').one()
            authority2 = cnx.create_entity('Organization', name=u'boss club')
            user = self.create_user(cnx, login=u'user', groups=('users',))
            cnx.commit()
            # jdoe is in authority2 and user has no authority: OK
            jdoe = testutils.agent(cnx, u'jdoe', authority=authority2)
            jdoe.cw_set(agent_user=user)
            cnx.commit()
            # jdoe is in authority2 and attempt to set user in authority1: KO
            with self.assertValidationError(cnx):
                user.cw_set(authority=authority1)
                cnx.commit()
            cnx.rollback()
            # break link between agent and user
            jdoe.cw_set(agent_user=None)
            cnx.commit()
            user.cw_set(authority=authority1)
            # jdoe is in authority2, user in authority1 and attempt to link them KO
            with self.assertValidationError(cnx):
                jdoe.cw_set(agent_user=user)


class AuthorityRecordTC(CubicWebTC):
    assertUnauthorizedRQL = assertUnauthorizedRQL

    def test_fti(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.create_entity('AgentKind', name=u'Queen')
            agent = testutils.authority_record(cnx, u'Guenievre', kind=u'Queen',
                                               end_date=date(476, 02, 9))
            address = cnx.create_entity('PostalAddress', street=u"1 av. de l'europe",
                                        postalcode=u'31400', city=u'Toulouse')
            cnx.create_entity('AgentPlace', name=u'place', place_address=address, place_agent=agent)
            cnx.create_entity('AgentFunction', name=u'function', function_agent=agent)
            cnx.create_entity('LegalStatus', term=u'legal status', legal_status_agent=agent)
            cnx.commit()
            for search in (u'guenievre', u'europe', u'place', u'function', u'legal status'):
                with self.subTest(search=search):
                    self.assertEqual(cnx.execute('AuthorityRecord X WHERE X has_text %(search)s',
                                                 {'search': search}).one().eid, agent.eid)

    def test_graph_structure(self):
        graph = PERMISSIONS_GRAPHS['AuthorityRecord'](self.schema)
        expected = {
            'AgentFunction': {('function_agent', 'subject'): set(['AuthorityRecord'])},
            'AgentPlace': {('place_agent', 'subject'): set(['AuthorityRecord'])},
            'Citation': {('has_citation', 'object'): set([
                'GeneralContext', 'Mandate', 'Occupation', 'AgentFunction',
                'AgentPlace', 'History', 'LegalStatus',
            ])},
            'EACResourceRelation': {('resource_relation_agent', 'subject'):
                                    set(['AuthorityRecord'])},
            'EACSource': {('source_agent', 'subject'): set(['AuthorityRecord'])},
            'GeneralContext': {('general_context_of', 'subject'): set(['AuthorityRecord'])},
            'History': {('history_agent', 'subject'): set(['AuthorityRecord'])},
            'LegalStatus': {('legal_status_agent', 'subject'): set(['AuthorityRecord'])},
            'Mandate': {('mandate_agent', 'subject'): set(['AuthorityRecord'])},
            'Occupation': {('occupation_agent', 'subject'): set(['AuthorityRecord'])},
            'PostalAddress': {('place_address', 'object'): set(['AgentPlace'])},
            'Structure': {('structure_agent', 'subject'): set(['AuthorityRecord'])},
        }
        struct = dict(
            (k, dict((rel, set(targets)) for rel, targets in v.items()))
            for k, v in graph.parent_structure('AuthorityRecord').items())
        self.assertEqual(struct, expected)

    def test_optional_relations(self):
        graph = PERMISSIONS_GRAPHS['AuthorityRecord'](self.schema)
        structure = graph.parent_structure('AuthorityRecord')
        opts = optional_relations(self.schema, structure)
        expected = {'Citation': set([('has_citation', 'object')])}
        self.assertEqual(opts, expected)

    def test_relations_consistency(self):
        graph = PERMISSIONS_GRAPHS['AuthorityRecord'](self.schema)
        structure = graph.parent_structure('AuthorityRecord')
        structurals, optionals, mandatories = graph_relations(
            self.schema, structure)
        self.assertEqual(structurals - optionals, mandatories)


class ConceptSchemeTC(CubicWebTC):

    def test_graph_structure(self):
        graph = PERMISSIONS_GRAPHS['ConceptScheme'](self.schema)
        expected = {
            'Concept': {('in_scheme', 'subject'): ['ConceptScheme']},
            'Label': {('label_of', 'subject'): ['Concept']},
        }
        self.assertEqual(graph.parent_structure('ConceptScheme'),
                         expected)

    def test_optional_relations(self):
        graph = PERMISSIONS_GRAPHS['ConceptScheme'](self.schema)
        opts = optional_relations(self.schema,
                                  graph.parent_structure('ConceptScheme'))
        expected = {}
        self.assertEqual(opts, expected)

    def test_relations_consistency(self):
        graph = PERMISSIONS_GRAPHS['ConceptScheme'](self.schema)
        structure = graph.parent_structure('ConceptScheme')
        structurals, optionals, mandatories = graph_relations(
            self.schema, structure)
        self.assertEqual(structurals - optionals, mandatories)


class SEDAProfileTC(CubicWebTC):
    assertValidationError = assertValidationError

    def test_fti(self):
        # "Reverse" text to be searched in order not to be troubled by other
        # entities that may live in the DB (e.g. Concepts) with similar text.
        with self.admin_access.client_cnx() as cnx:
            profile = testutils.setup_seda_profile(cnx)
            archive = profile.archives[0]
            archive.name.cw_set(value=u'profile name'[::-1])
            ac = archive.access_restriction_code
            ac.cw_set(user_annotation=u'some annotation'[::-1])
            testutils.seda_scheme(cnx, u'seda_document_type_code', u'preferred', u'CDO')
            seda_document_type_code_value = cnx.execute(
                'Concept X WHERE X preferred_label L, L label "CDO"').one()
            seda_document_type_code = cnx.create_entity(
                'SEDADocumentTypeCode',
                seda_document_type_code_value=seda_document_type_code_value)
            cnx.create_entity('ProfileDocument',
                              seda_description=cnx.create_entity(
                                  'SEDADescription',
                                  value=u'document description'[::-1]),
                              seda_document_type_code=seda_document_type_code,
                              seda_parent=archive)
            cnx.commit()
            for search in ('name', 'annotation', 'description'):
                with self.subTest(search=search):
                    rset = cnx.execute('Any X WHERE X has_text %(search)s',
                                       {'search': search[::-1]})
                    self.assertEqual([r for r, in rset.rows], [profile.eid])

    def test_graph_structure(self):
        graph = PERMISSIONS_GRAPHS['ProfileArchiveObject'](self.schema)
        expected = {
            'ProfileArchiveObject': {
                ('seda_parent', 'subject'): [
                    'SEDAProfile', 'ProfileArchiveObject']},
            'ProfileDocument': {
                ('seda_parent', 'subject'): ['ProfileArchiveObject']},
            'SEDAAccessRestrictionCode': {
                ('seda_access_restriction_code', 'object'): [
                    'ProfileArchiveObject']},
            'SEDAAppraisalRule': {
                ('seda_appraisal_rule', 'object'): ['ProfileArchiveObject']},
            'SEDAAppraisalRuleCode': {
                ('seda_appraisal_rule_code', 'object'): ['SEDAAppraisalRule']},
            'SEDAAppraisalRuleDuration': {
                ('seda_appraisal_rule_duration', 'object'): [
                    'SEDAAppraisalRule']},
            'SEDACharacterSetCode': {
                ('seda_character_set_code', 'object'): ['ProfileDocument']},
            'SEDAContentDescription': {
                ('seda_content_description', 'object'): [
                    'ProfileArchiveObject']},
            'SEDADate': {
                ('seda_latest_date', 'object'): ['SEDAContentDescription'],
                ('seda_oldest_date', 'object'): ['SEDAContentDescription']},
            'SEDADescription': {
                ('seda_description', 'object'): ['SEDAContentDescription',
                                                 'ProfileDocument']},
            'SEDADescriptionLevel': {
                ('seda_description_level', 'object'): [
                    'SEDAContentDescription']},
            'SEDADocumentTypeCode': {
                ('seda_document_type_code', 'object'): ['ProfileDocument']},
            'SEDAFileTypeCode': {
                ('seda_file_type_code', 'object'): ['ProfileDocument']},
            'SEDAKeyword': {
                ('seda_keyword_of', 'subject'): ['SEDAContentDescription']},
            'SEDAName': {
                ('seda_name', 'object'): ['ProfileArchiveObject']},
        }
        self.assertEqual(graph.parent_structure('SEDAProfile'),
                         expected)

    def test_optional_relations(self):
        graph = PERMISSIONS_GRAPHS['ProfileArchiveObject'](self.schema)
        opts = optional_relations(self.schema,
                                  graph.parent_structure('SEDAProfile'))
        expected = {
            'ProfileArchiveObject': set([('seda_parent', 'subject')]),
            'ProfileDocument': set([('seda_parent', 'subject')]),
            'SEDADate': set([('seda_oldest_date', 'object'),
                             ('seda_latest_date', 'object')]),
        }
        self.assertEqual(opts, expected)

    def test_relations_consistency(self):
        graph = PERMISSIONS_GRAPHS['ProfileArchiveObject'](self.schema)
        structure = graph.parent_structure('SEDAProfile')
        structurals, optionals, mandatories = graph_relations(
            self.schema, structure)
        self.assertEqual(structurals - optionals, mandatories)

    def _test_relative_permissions_through(self, permission):
        actual = [e.expression
                  for e in self.schema['ProfileArchiveObject'].permissions[permission]]
        expected = [
            u'NOT EXISTS(X seda_parent A), U in_group G, G name IN("managers", "users")',
            u'X seda_parent A, U has_update_permission A',
        ]
        self.assertEqual(actual, expected)
        actual = [e.expression
                  for e in self.schema['SEDADate'].permissions[permission]]
        expected = [
            'A seda_oldest_date X, U has_update_permission A',
            'A seda_latest_date X, U has_update_permission A',
        ]
        self.assertEqual(actual, expected)

    def test_relative_update_permissions_through(self):
        self._test_relative_permissions_through('update')

    def test_relative_delete_permissions_through(self):
        self._test_relative_permissions_through('delete')

    def test_relative_permissions_of(self):
        rdef = self.schema['seda_parent'].rdef('ProfileDocument', 'ProfileArchiveObject')
        rqlexpr, = rdef.permissions['add']
        self.assertEqual(rqlexpr.expression, 'U has_update_permission O')
        rdef = self.schema['seda_oldest_date'].rdef('SEDAContentDescription', 'SEDADate')
        rqlexpr, = rdef.permissions['delete']
        self.assertEqual(rqlexpr.expression, 'U has_update_permission S')

    def test_dates_constraints(self):
        with self.admin_access.client_cnx() as cnx:
            profile = testutils.setup_seda_profile(cnx)
            archive = profile.archives[0]
            desc = archive.content_description
            date = cnx.create_entity('SEDADate',
                                     reverse_seda_oldest_date=desc)
            cnx.commit()

            def assert_error(cm, date_type):
                self.assertEqual(cm.exception.errors,
                                 {'seda_%s_date-object' % date_type:
                                  u'SEDA date cannot be used as oldest and latest'})

            datetype = 'latest'
            with self.subTest(datetype=datetype):
                with self.assertValidationError(cnx) as cm:
                    desc.cw_set(seda_latest_date=date)
                assert_error(cm, datetype)
                desc.cw_set(seda_oldest_date=None)
                desc.cw_set(seda_latest_date=date)
                cnx.commit()

            datetype = 'oldest'
            with self.subTest(datetype=datetype):
                with self.assertValidationError(cnx) as cm:
                    desc.cw_set(seda_oldest_date=date)
                assert_error(cm, datetype)


class SEDAKeywordTC(CubicWebTC):
    assertValidationError = assertValidationError

    def test_scheme_value_constraint(self):
        with self.admin_access.client_cnx() as cnx:
            profile = testutils.setup_seda_profile(cnx)
            cnx.commit()
            cd = profile.archives[0].content_description
            naa = testutils.naa(cnx)
            unrelated_scheme = cnx.create_entity('ConceptScheme', ark_naa=naa)
            scheme = cnx.create_entity('ConceptScheme', ark_naa=naa)
            concept = scheme.add_concept(u'this')
            cnx.commit()
            with self.assertValidationError(cnx) as cm:
                kw = cnx.create_entity('SEDAKeyword', seda_keyword_of=cd)
                kw.cw_set(seda_keyword_value=concept)
                kw.cw_set(seda_keyword_scheme=unrelated_scheme)
            self.assertEqual(cm.exception.errors,
                             {'seda_keyword_value-subject':
                              u"concept doesn't belong to the scheme associated to the keyword"})

            kw = cnx.create_entity('SEDAKeyword', seda_keyword_of=cd)
            kw.cw_set(seda_keyword_value=concept,
                      seda_keyword_scheme=scheme)
            cnx.commit()


class SecurityTC(CubicWebTC):
    """Test case for permissions set in the schema"""
    assertUnauthorized = assertUnauthorized

    def test_agentkind_type(self):
        with self.admin_access.cnx() as cnx:
            kind = cnx.find('AgentKind', name=u'person').one()
            # no one can update nor delete a kind
            with self.assertRaises(Unauthorized):
                kind.cw_set(name=u'gloups')
                cnx.commit()
            cnx.rollback()
            with self.assertRaises(Unauthorized):
                kind.cw_delete()
                cnx.commit()
            cnx.rollback()

    def test_agent_kind_relation(self):
        """Test we can only change kind from unknown to another."""
        with self.admin_access.repo_cnx() as cnx:
            agent = testutils.authority_record(cnx, u'bob', kind=u'unknown-agent-kind')
            cnx.commit()
            agent.cw_set(agent_kind=cnx.find('AgentKind', name=u'person').one())
            cnx.commit()
            with self.assertRaises(Unauthorized):
                agent.cw_set(agent_kind=cnx.find('AgentKind', name=u'authority').one())

    def test_authority_type(self):
        with self.admin_access.repo_cnx() as cnx:
            self.create_user(cnx, login=u'user', groups=('users',))
            cnx.commit()
        with self.new_access('user').client_cnx() as cnx:
            with self.assertUnauthorized(cnx):
                cnx.create_entity('Organization', name=u'dream team')

    def test_authority_relation(self):
        with self.admin_access.repo_cnx() as cnx:
            self.create_user(cnx, login=u'user', groups=('users',),
                             authority=cnx.find('Organization').one())
            agent = testutils.agent(cnx, u'user')
            authority = cnx.create_entity('Organization', name=u'dream team')
            cnx.commit()
            # even manager can't change an agent's authority
            with self.assertUnauthorized(cnx):
                agent.cw_set(authority=authority.eid)
        with self.new_access('user').client_cnx() as cnx:
            agent = cnx.entity_from_eid(agent.eid)
            # user can't change its own authority
            with self.assertUnauthorized(cnx):
                agent.cw_set(authority=authority.eid)
            # user can't create an agent in another authority than its own
            with self.assertUnauthorized(cnx):
                testutils.agent(cnx, u'new agent', authority=authority.eid)
            # though he can add an agent to its own authority
            testutils.agent(cnx, u'new agent')
            cnx.commit()

    def test_agent_user(self):
        with self.admin_access.repo_cnx() as cnx:
            user1 = self.create_user(cnx, login=u'user1', groups=('users',),
                                     authority=cnx.find('Organization').one())
            user2 = self.create_user(cnx, login=u'user2', groups=('users',),
                                     authority=cnx.find('Organization').one())
            agent = testutils.agent(cnx, u'user1', agent_user=user1)
            cnx.commit()
        with self.new_access('user1').client_cnx() as cnx:
            agent = cnx.entity_from_eid(agent.eid)
            # user can't change its own user
            with self.assertUnauthorized(cnx):
                agent.cw_set(agent_user=user2.eid)
            with self.assertUnauthorized(cnx):
                agent.cw_set(agent_user=None)
            # user can't create an agent and specify its associated user
            with self.assertUnauthorized(cnx):
                testutils.agent(cnx, u'user2', agent_user=user2.eid)
            agent2 = testutils.agent(cnx, u'user2')
            cnx.commit()
            with self.assertUnauthorized(cnx):
                agent2.cw_set(agent_user=user2.eid)

    def test_authority_record_base(self):
        with self.admin_access.repo_cnx() as cnx:
            self.create_user(cnx, login=u'toto', groups=('users', 'guests'),
                             authority=cnx.find('Organization').one())
            cnx.commit()
        with self.new_access('toto').client_cnx() as cnx:
            function = cnx.create_entity('AgentFunction', name=u'grouillot')
            testutils.authority_record(cnx, u'bob', reverse_function_agent=function)
            cnx.commit()
        with self.admin_access.repo_cnx() as cnx:
            cnx.execute('DELETE U in_group G WHERE U login "toto", G name "users"')
            cnx.commit()
        with self.new_access('toto').client_cnx() as cnx:
            agent = cnx.find('AuthorityRecord', name=u'bob').one()
            # guest user can't modify an authority record
            with self.assertUnauthorized(cnx):
                agent.cw_set(name=u'bobby')
            with self.assertUnauthorized(cnx):
                agent.reverse_function_agent[0].cw_set(name=u'director')

    def test_authority_record_wf_permissions(self):
        with self.admin_access.repo_cnx() as cnx:
            self.create_user(cnx, login=u'toto', groups=('users', 'guests'),
                             authority=cnx.find('Organization').one())
            cnx.commit()
        with self.new_access('toto').client_cnx() as cnx:
            function = cnx.create_entity('AgentFunction', name=u'grouillot')
            agent = testutils.authority_record(cnx, u'bob', reverse_function_agent=function)
            cnx.commit()
            iwf = agent.cw_adapt_to('IWorkflowable')
            iwf.fire_transition('publish')
            cnx.commit()
            # we can still modify a published agent
            agent.cw_set(name=u'bobby')
            function.cw_set(name=u'director')
            cnx.commit()

    def test_update_root_badgroup(self):
        with self.admin_access.repo_cnx() as cnx:
            self.create_user(cnx, login=u'toto', groups=('users', 'guests'))
            cnx.commit()
        with self.new_access('toto').client_cnx() as cnx:
            testutils.publishable_profile(cnx, title=u'pp')
            cnx.commit()
        with self.admin_access.repo_cnx() as cnx:
            cnx.execute(
                'DELETE U in_group G WHERE U login "toto", G name "users"')
            cnx.commit()
        with self.new_access('toto').client_cnx() as cnx:
            profile = cnx.find('SEDAProfile', title=u'pp').one()
            with self.assertUnauthorized(cnx):
                profile.cw_set(title=u'qq')

    def test_update_child(self):
        with self.admin_access.repo_cnx() as cnx:
            self.create_user(cnx, login=u'toto', groups=('users', 'guests'),
                             authority=cnx.find('Organization').one())
            cnx.commit()
        with self.new_access('toto').client_cnx() as cnx:
            record = testutils.authority_record(cnx, u'bob')
            function = cnx.create_entity('AgentFunction', name=u'boss',
                                         function_agent=record)
            cnx.commit()
            # Draft -> OK.
            function.cw_set(name=u'grouyo')
            cnx.commit()
            cnx.create_entity('GeneralContext', content=u'plop',
                              general_context_of=record)
            cnx.commit()
            iwf = record.cw_adapt_to('IWorkflowable')
            iwf.fire_transition('publish')
            cnx.commit()
            # Published -> still OK for record.
            record.cw_set(name=u'big boss')
            cnx.commit()
            cnx.create_entity('History', text=u'yellow sponge',
                              history_agent=record)
            cnx.commit()

    def test_sedaprofile_wf_permissions(self):
        with self.admin_access.repo_cnx() as cnx:
            profile = testutils.publishable_profile(cnx, title=u'pp')
            name = cnx.create_entity('SEDAName')
            profile_archive_obj = cnx.create_entity(
                'ProfileArchiveObject', seda_name=name, seda_parent=profile)
            cnx.commit()
            # Profile in draft, modifications are allowed.
            profile.cw_set(title=u'ugh')
            cnx.commit()
            name.cw_set(value=u'argh')
            cnx.commit()
            seda_date = cnx.create_entity('SEDADate')
            profile_archive_obj.content_description.cw_set(seda_oldest_date=seda_date)
            cnx.commit()
            # Profile published, no modification allowed.
            profile.cw_adapt_to('IWorkflowable').fire_transition('publish')
            cnx.commit()
            with self.assertUnauthorized(cnx):
                profile.cw_set(title=u'gloups')
            with self.assertUnauthorized(cnx):
                name.cw_set(value=u'zorglub')
            with self.assertUnauthorized(cnx):
                seda_date.cw_set(value=date.today())
            with self.assertUnauthorized(cnx):
                profile_archive_obj.content_description.cw_set(
                    seda_latest_date=cnx.create_entity('SEDADate'))
            with self.assertUnauthorized(cnx):
                seda_date.cw_delete()
            with self.assertUnauthorized(cnx):
                profile_archive_obj.cw_delete()

    def test_sedaprofile_no_parent_wf_permissions(self):
        with self.admin_access.repo_cnx() as cnx:
            name = cnx.create_entity('SEDAName')
            profile_archive_obj = cnx.create_entity(
                'ProfileArchiveObject', seda_name=name)
            cnx.commit()
            # document unit has no parent, modifications are allowed.
            name.cw_set(value=u'argh')
            cnx.commit()
        with self.new_access('anon').client_cnx() as cnx:
            name = cnx.entity_from_eid(name.eid)
            profile_archive_obj = cnx.entity_from_eid(profile_archive_obj.eid)
            with self.assertUnauthorized(cnx):
                name.cw_set(value=u'zorglub')
            with self.assertUnauthorized(cnx):
                profile_archive_obj.cw_set(user_annotation=u'zorglub')
            ar = cnx.create_entity('SEDAAccessRestrictionCode')
            profile_archive_obj.cw_set(seda_access_restriction_code=ar)
            with self.assertUnauthorized(cnx):
                pass  # exception raised by the commit done by the context manager
        with self.admin_access.repo_cnx() as cnx:
            profile_archive_obj = cnx.entity_from_eid(profile_archive_obj.eid)
            profile_archive_obj.cw_delete()
            cnx.commit()

    def test_conceptscheme_wf_permissions(self):
        with self.admin_access.cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', ark_naa=testutils.naa(cnx))
            cnx.commit()
            # in draft, modifications are allowed.
            concept = scheme.add_concept(u'blah')
            cnx.commit()
            # published, can't modify existing content.
            scheme.cw_adapt_to('IWorkflowable').fire_transition('publish')
            cnx.commit()
            with self.assertUnauthorized(cnx):
                scheme.cw_set(description=u'plop')
            with self.assertUnauthorized(cnx):
                concept.preferred_label[0].cw_set(label=u'plop')
            # though addition of new concepts / labels is fine
            new_concept = scheme.add_concept(u'plop')
            cnx.commit()
            new_label = cnx.create_entity('Label', label=u'arhg', label_of=concept)
            cnx.commit()
            # while deletion is fine for label but not for concept nor scheme
            new_label.cw_delete()
            cnx.commit()
            with self.assertUnauthorized(cnx):
                scheme.cw_delete()
            with self.assertUnauthorized(cnx):
                new_concept.cw_delete()


if __name__ == '__main__':
    unittest.main()



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref views for organization-like entities (Organization, OrganizationUnit and
Agent types)
"""

from cubicweb import tags
from cubicweb.predicates import partial_has_related_entities, is_instance, match_kwargs
from cubicweb.view import EntityView
from cubicweb.uilib import js
from cubicweb.web import formwidgets as fw
from cubicweb.web.views import ajaxcontroller, tabs, uicfg

from cubes.relationwidget import views as rwdg

from . import (SubviewsTabView, RelatedEntitiesListView,
               configure_relation_widget, dropdown_button, has_rel_perm)

_ = unicode

affk = uicfg.autoform_field_kwargs
afs = uicfg.autoform_section
abaa = uicfg.actionbox_appearsin_addmenu
pvs = uicfg.primaryview_section
pvds = uicfg.primaryview_display_ctrl
rdc = uicfg.reledit_ctrl


def authority_form_value(form, field):
    """Field default value function returning the current user's authority or None"""
    if form._cw.user.authority:
        return (unicode(form._cw.user.authority[0].eid),)
    return []

pvs.tag_subject_of(('*', 'authority', '*'), 'attributes')
afs.tag_subject_of(('*', 'authority', '*'), 'main', 'attributes')
affk.tag_subject_of(('*', 'authority', '*'), {'value': authority_form_value})

pvs.tag_subject_of(('*', 'authority_record', '*'), 'attributes')
afs.tag_subject_of(('*', 'authority_record', '*'), 'main', 'attributes')
affk.tag_subject_of(('*', 'authority_record', '*'),
                    {'widget': rwdg.RelationFacetWidget(
                        dialog_options={'width': 800},
                        no_creation_form=True)})

for etype, attr in (
    ('Organization', 'ark'),
    ('Organization', 'name'),
    ('OrganizationUnit', 'ark'),
    ('OrganizationUnit', 'name'),
    ('Agent', 'ark'),
    ('Agent', 'name'),
):
    affk.set_field_kwargs(etype, attr, widget=fw.TextInput({'size': 80}))
    if attr == 'ark':
        affk.set_field_kwargs(etype, attr, required=False)


# Organization

pvs.tag_object_of(('*', 'authority', 'Organization'), 'hidden')
afs.tag_object_of(('*', 'authority', 'Organization'), 'main', 'hidden')
abaa.tag_object_of(('*', 'authority', 'Organization'), False)
pvs.tag_subject_of(('Organization', 'archival_unit', '*'), 'attributes')
pvs.tag_subject_of(('Organization', 'ark_naa', 'ArkNameAssigningAuthority'), 'attributes')
afs.tag_subject_of(('Organization', 'ark_naa', 'ArkNameAssigningAuthority'), 'main', 'attributes')


class OrganizationTabbedPrimaryView(tabs.TabbedPrimaryView):
    """Tabbed primary view for Organization entity type."""
    __select__ = tabs.TabbedPrimaryView.__select__ & is_instance('Organization')
    tabs = [
        'main_tab',
        _('saem_org_orgunit_tab'),
        _('saem_org_agent_tab'),
    ]


class OrganizationUnitTab(SubviewsTabView):
    __regid__ = 'saem_org_orgunit_tab'
    __select__ = SubviewsTabView.__select__ & is_instance('Organization')
    rtype_role_targets = [
        ('authority', 'object', 'OrganizationUnit'),
    ]
    subvids = (
        'saem.org.organizationunits',
    )


class AgentTab(OrganizationUnitTab):
    __regid__ = 'saem_org_agent_tab'
    rtype_role_targets = [
        ('authority', 'object', 'Agent'),
    ]
    subvids = (
        'saem.org.agents',
    )


class OrganizationRelatedOrganizationUnitListView(RelatedEntitiesListView):
    __regid__ = 'saem.org.organizationunits'
    rtype = 'authority'
    role = 'object'
    target_etype = 'OrganizationUnit'

    _('creating Agent (Agent authority Organization %(linkto)s)')

    @property
    def title(self):
        return None


class OrganizationRelatedAgentListView(OrganizationRelatedOrganizationUnitListView):
    __regid__ = 'saem.org.agents'
    target_etype = 'Agent'

    _('creating OrganizationUnit (OrganizationUnit authority Organization %(linkto)s)')


# OrganizationUnit

pvs.tag_subject_of(('OrganizationUnit', 'archival_role', '*'), 'attributes')
afs.tag_subject_of(('OrganizationUnit', 'archival_role', '*'), 'main', 'attributes')
afs.tag_object_of(('*', 'archival_unit', 'OrganizationUnit'), 'main', 'hidden')
# archival_role is used in constraints for archival_agent and seda_transferring_agent relations
rdc.tag_attribute(('OrganizationUnit', 'archival_role'), {'reload': True})
pvs.tag_subject_of(('OrganizationUnit', 'contact_point', '*'), 'attributes')
afs.tag_subject_of(('OrganizationUnit', 'contact_point', '*'), 'main', 'attributes')

pvs.tag_subject_of(('*', 'seda_transferring_agent', '*'), 'hidden')
afs.tag_subject_of(('*', 'seda_transferring_agent', '*'), 'main', 'hidden')
pvs.tag_object_of(('*', 'seda_transferring_agent', '*'), 'hidden')
afs.tag_object_of(('*', 'seda_transferring_agent', '*'), 'main', 'hidden')
abaa.tag_object_of(('*', 'seda_transferring_agent', '*'), False)

pvs.tag_subject_of(('*', 'related_concept_scheme', '*'), 'hidden')
afs.tag_subject_of(('*', 'related_concept_scheme', '*'), 'main', 'hidden')
pvs.tag_object_of(('*', 'related_concept_scheme', '*'), 'hidden')
afs.tag_object_of(('*', 'related_concept_scheme', '*'), 'main', 'hidden')
abaa.tag_object_of(('*', 'related_concept_scheme', '*'), False)


class OrganizationUnitTabbedPrimaryView(tabs.TabbedPrimaryView):
    """Tabbed primary view for OrganizationUnit"""
    __select__ = tabs.TabbedPrimaryView.__select__ & is_instance('OrganizationUnit')
    tabs = [
        'main_tab',
        _('saem_ref_deposit_ou_concepts_profiles'),
    ]


class DepositOrganizationUnitConceptsProfilesTab(SubviewsTabView):
    """SEDA profiles and concept schemes used by an agent with role "deposit"."""
    __regid__ = 'saem_ref_deposit_ou_concepts_profiles'
    __select__ = EntityView.__select__ & is_instance('OrganizationUnit')
    subvids = (
        'saem.ou.schemes',
        'saem.ou.profiles',
    )

    def entity_call(self, entity):
        self.w(u'<div id="%s%s">' % (self.__regid__, entity.eid))
        if (not entity.has_role('deposit')
                and has_rel_perm('add', entity, 'seda_transferring_agent', 'object')):
            msg = self._cw._("You can't add profiles or concept schemes to agent without the "
                             "'deposit' role.")
            self.w('<div class="alert alert-warning">%s</div>' % msg)
        else:
            self.generate_add_button(entity)
        super(DepositOrganizationUnitConceptsProfilesTab, self).entity_call(entity)
        self.w(u'</div>')

    def generate_add_button(self, entity):
        """Add html for the add button on the right hand corner of the tab, if the logged user may
        add some relation.
        """
        divid = "relatedentities%s" % entity.eid
        links = []
        for etype, rtype, role in (('SEDAProfile', 'seda_transferring_agent', 'object'),
                                   ('ConceptScheme', 'related_concept_scheme', 'subject')):
            if not has_rel_perm('add', entity, rtype, role, target_etype=etype):
                continue
            relation = '%s:%s:%s' % (rtype, etype, role)
            search_url = self._cw.build_url('ajax', fname='view', vid='search_related_entities',
                                            eid=entity.eid,
                                            __modal=1, multiple='1', relation=relation)
            title = (self._cw._('Search %s to link to the agent')
                     % self._cw.__(etype + '_plural').lower())
            validate = js.saem.buildRelationValidate(entity.eid, rtype, role, self.__regid__)
            url = configure_relation_widget(self._cw, divid, search_url, title, True, validate)
            links.append(tags.a(self._cw._(etype), href=url, klass=''))
        if links:
            rwdg.boostrap_dialog(self.w, self._cw._, divid, u'')
            self.w(tags.div(id=divid, style='display: none'))
            self.w(dropdown_button(self._cw._('add'), *links))
            self.w(tags.div(klass='clearfix'))


class OrganizationUnitSearchForRelatedEntitiesView(rwdg.SearchForRelatedEntitiesView):
    __select__ = (rwdg.SearchForRelatedEntitiesView.__select__
                  & (rwdg.edited_relation('seda_transferring_agent')
                     | rwdg.edited_relation('related_concept_scheme')))
    has_creation_form = False

    def linkable_rset(self):
        """Return rset of entities to be displayed as possible values for the edited relation. You
        may want to override this.
        """
        entity = self.compute_entity()
        rtype, tetype, role = self._cw.form['relation'].split(':')
        return entity.unrelated(rtype, tetype, role, ordermethod='fetch_order')


# cw provide delete_relation and add_relation. Implements add_relations because it's easy but we
# could do it using several call to add_relation
@ajaxcontroller.ajaxfunc
def add_relations(self, eid, rtype, role, related_eids):
    """Add relation `rtype` between `eid` with role `role` and `related_eid`."""
    rql = 'SET S {rtype} O WHERE S eid %(eids)s, O eid %(eido)s'.format(rtype=rtype)
    for related_eid in related_eids:
        self._cw.execute(rql, {'eids': eid if role == 'subject' else int(related_eid),
                               'eido': eid if role == 'object' else int(related_eid)})


class RTypeListView(EntityView):
    """List of related entities the agent is using"""
    __abstract__ = True
    __select__ = EntityView.__select__ & partial_has_related_entities()
    rtype = None
    role = None

    def entity_call(self, entity, **kwargs):
        title = (self._cw.__(self.rtype + '_object')
                 if self.role == 'object' else self._cw.__(self.rtype))
        self.w(tags.h2(self._cw._(title)))
        rset = entity.related(self.rtype, self.role)
        subvid = 'saem.listitem'
        self._cw.view('list', rset=rset, w=self.w, subvid=subvid, unit=entity,
                      rtype=self.rtype, role=self.role, __redirectpath=entity.rest_path())


class OrganizationUnitUsingConceptSchemeListView(RTypeListView):
    """View for ConceptScheme, to be displayed in the context of an Agent"""
    __regid__ = 'saem.ou.schemes'
    rtype = 'related_concept_scheme'
    role = 'subject'


class RelatedSEDAProfileListView(RTypeListView):
    """View for SEDAProfile, to be displayed in the context of an Agent"""
    __regid__ = 'saem.ou.profiles'
    rtype = 'seda_transferring_agent'
    role = 'object'


class OrganizationUnitUsingListItemView(EntityView):
    """Extended 'oneline' view for entities related to an OrganizationUnit, including link to remove
    the relation.
    """
    __regid__ = 'saem.listitem'
    __select__ = EntityView.__select__ & match_kwargs('unit', 'rtype', 'role')

    # XXX usually expect role to be the role of the entity, here it's the role of the agent
    def entity_call(self, entity, unit, rtype, role, **editurlparams):
        entity.view('outofcontext', w=self.w)
        if has_rel_perm('delete', unit, rtype, role, target_entity=entity):
            self._cw.add_js(('cubicweb.ajax.js', 'cubes.saem_ref.js'))
            self.w(u'<div class="pull-right">')
            jscall = js.saem.ajaxRemoveRelation(unit.eid, entity.eid, rtype, role,
                                                'saem_ref_deposit_ou_concepts_profiles')
            self.w(tags.a(title=self._cw._('delete'), klass='icon-trash',
                          href='javascript: %s' % jscall))
            self.w(u'</div>')


# Agent
pvs.tag_subject_of(('Agent', 'phone_number', '*'), 'attributes')
afs.tag_subject_of(('Agent', 'phone_number', '*'), 'main', 'inlined')
pvds.tag_subject_of(('Agent', 'use_email', '*'), {'vid': 'text'})

afs.tag_object_of(('*', 'contact_point', 'Agent'), 'main', 'hidden')

pvs.tag_object_of(('*', 'agent_user', '*'), 'hidden')
pvs.tag_subject_of(('Agent', 'agent_user', '*'), 'hidden')
afs.tag_subject_of(('Agent', 'agent_user', '*'), 'main', 'hidden')



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem_ref common test tools"""

import unittest
from doctest import Example

from lxml import etree
from lxml.doctestcompare import LXMLOutputChecker

from cubicweb import NoResultError


def agent(cnx, name, **kwargs):
    """Return an Agent with specified name."""
    if 'authority' not in kwargs:
        authority = cnx.find('Organization', name=u'Default authority').one()
        kwargs['authority'] = authority
    else:
        authority = kwargs['authority']
        if isinstance(authority, int):
            authority = cnx.entity_from_eid(authority)
    if not authority.ark_naa:
        with cnx.security_enabled(False, False):
            authority.cw_set(ark_naa=naa(cnx))
    return cnx.create_entity('Agent', name=name, **kwargs)


def organization_unit(cnx, name, archival_roles=(), **kwargs):
    """Return an OrganizationUnit with specified name and archival roles."""
    if 'authority' not in kwargs:
        authority = cnx.find('Organization', name=u'Default authority').one()
        kwargs['authority'] = authority
    else:
        authority = kwargs['authority']
        if isinstance(authority, int):
            authority = cnx.entity_from_eid(authority)
    if not authority.ark_naa:
        with cnx.security_enabled(False, False):
            authority.cw_set(ark_naa=naa(cnx))
    roles_eid = [cnx.find('ArchivalRole', name=role)[0][0] for role in archival_roles]
    return cnx.create_entity('OrganizationUnit', name=name,
                             archival_role=roles_eid, **kwargs)


def authority_record(cnx, name, kind=u'person', **kwargs):
    """Return an AuthorityRecord with specified kind and name."""
    kind_eid = cnx.find('AgentKind', name=kind)[0][0]
    if 'ark_naa' not in kwargs:
        authority = cnx.find('Organization', name=u'Default authority').one()
        if not authority.ark_naa:
            with cnx.security_enabled(False, False):
                authority.cw_set(ark_naa=naa(cnx))
        kwargs['ark_naa'] = authority.ark_naa
    return cnx.create_entity('AuthorityRecord', name=name,
                             agent_kind=kind_eid, **kwargs)


def naa(cnx):
    try:
        return cnx.find('ArkNameAssigningAuthority').one()
    except NoResultError:
        return cnx.create_entity('ArkNameAssigningAuthority', who=u'TEST', what=0)


def authority_with_naa(cnx):
    authority = cnx.find('Organization', name=u'Default authority').one()
    if not authority.ark_naa:
        authority.cw_set(ark_naa=naa(cnx))
    return authority


def setup_scheme(cnx, title, *labels):
    """Return info new concept scheme"""
    scheme = cnx.create_entity('ConceptScheme', title=title, ark_naa=naa(cnx))
    for label in labels:
        scheme.add_concept(label)
    return scheme


def setup_seda_agents(cnx):
    """Return two OrganizationUnits, one with the 'deposit' archival role,
    another with the 'archival' archival role.
    """
    alice = agent(cnx, u'alice')
    city = organization_unit(cnx, u'New York city',
                             archival_roles=[u'archival'], contact_point=alice)
    cnx.create_entity('Organization', name=u'State of New York',
                      archival_unit=city)
    bob = organization_unit(cnx, u'bob', archival_roles=['deposit'])
    return bob, city


def setup_seda_profile(cnx, complete_archive=True, **kwargs):
    """Return a minimal SEDA profile with agents and a profile archive (with or without content
    description and access restriction code according to the `complete_archive` flag).
    """
    if 'seda_transferring_agent' not in kwargs:
        transferring_agent, _ = setup_seda_agents(cnx)
        kwargs['seda_transferring_agent'] = transferring_agent
    create_entity = cnx.create_entity
    profile = create_entity('SEDAProfile', ark_naa=naa(cnx), **kwargs)
    kwargs = {}
    if complete_archive:
        kwargs['seda_content_description'] = create_entity(
            'SEDAContentDescription',
            seda_description_level=create_entity('SEDADescriptionLevel'))
        kwargs['seda_access_restriction_code'] = create_entity('SEDAAccessRestrictionCode')
    create_entity('ProfileArchiveObject', seda_name=create_entity('SEDAName'),
                  seda_parent=profile, **kwargs)
    return profile


def publishable_profile(cnx, **kwargs):
    """Return a minimal SEDA profile that may be published."""
    create_entity = cnx.create_entity
    profile = create_entity('SEDAProfile', ark_naa=naa(cnx), **kwargs)
    create_entity('ProfileArchiveObject',
                  seda_parent=profile,
                  seda_name=create_entity('SEDAName'),
                  seda_content_description=create_entity(
                      'SEDAContentDescription',
                      seda_description_level=create_entity('SEDADescriptionLevel')),
                  seda_access_restriction_code=create_entity('SEDAAccessRestrictionCode'))
    return profile


def concept(cnx, label):
    """Return concept entity with the given preferred label (expected to be unique)."""
    return cnx.execute('Concept X WHERE X preferred_label L, L label %(label)s',
                       {'label': label}).one()


def seda_scheme(cnx, rtype, label_kind, *labels, **kwargs):
    """Build a ConceptScheme with `scheme_relation` pointing to `rtype` along
    with concepts with `labels` of `label_kind`.
    """
    rtype_eid = cnx.find('CWRType', name=rtype)[0][0]
    scheme = cnx.create_entity('ConceptScheme',
                               title=kwargs.get('scheme_title'),
                               scheme_relation=rtype_eid,
                               ark_naa=naa(cnx))
    for label in labels:
        scheme.add_concept(label, kind=label_kind)
    return scheme


#
# XML helpers
#

def xmlpp(string):
    """Parse and pretty-print XML data from `string`."""
    print etree.tostring(etree.fromstring(string), pretty_print=True)


class XmlTestMixin(unittest.TestCase):
    """Mixin class provinding additional assertion methods for checking XML data."""

    def assertXmlEqual(self, actual, expected):
        """Check that both XML strings represent the same XML tree."""
        checker = LXMLOutputChecker()
        if not checker.check_output(expected, actual, 0):
            message = checker.output_difference(Example("", expected), actual, 0)
            self.fail(message)

    def assertXmlValid(self, xml_data, xsd_filename, debug=False):
        """Validate an XML file (.xml) according to an XML schema (.xsd)."""
        with open(xsd_filename) as xsd:
            xmlschema = etree.XMLSchema(etree.parse(xsd))
        # Pretty-print xml_data to get meaningfull line information.
        xml_data = etree.tostring(etree.fromstring(xml_data), pretty_print=True)
        xml_data = etree.fromstring(xml_data)
        if debug and not xmlschema.validate(xml_data):
            xmlpp(xml_data)
        xmlschema.assertValid(xml_data)



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref tree views using jqTree"""

from functools import wraps

from cubicweb import tags, uilib
from cubicweb.view import EntityAdapter, EntityView, View
from cubicweb.predicates import adaptable, match_form_params
from cubicweb.web.views import ajaxcontroller, json


@ajaxcontroller.ajaxfunc(output_type='json')
def jqtree_reparent(self, child_eid, parent_eid):
    child = self._cw.entity_from_eid(child_eid)
    adapted = child.cw_adapt_to('IJQTree')
    if adapted is not None:
        return adapted.reparent(parent_eid)


class IJQTreeAdapter(EntityAdapter):
    """Adapt an entity into a JQTree node."""
    __regid__ = 'IJQTree'

    def maybe_parent_of(self):
        """Return a list of entity type which `entity` may be parent of."""
        return []

    def maybe_child(self):
        """Return True if entity may be child of another (i.e. may be
        reparented).
        """
        return False

    def reparent(self, peid):
        """Set entity as child of entity with `peid`."""
        pass

    def json(self, on_demand=False):
        """Return a JSON dict of an entity."""
        entity = self.entity
        label = entity.view('jqtree.label')
        data = {
            'label': label,
            'id': entity.eid,
            'type': entity.cw_etype,
            'maybeChild': self.maybe_child(),
            'maybeParentOf': self.maybe_parent_of(),
        }
        if on_demand:
            data['load_on_demand'] = True
        return data


def itree_adaptable(func):
    """Decorator for a function taking an entity as argument, checking that
    the entity is adaptable as ITree and eventually calling it or not.
    """
    @wraps(func)
    def wrapper(entity):
        itree = entity.cw_adapt_to('ITree')
        if itree:
            return func(entity)
        return ()
    return wrapper


@itree_adaptable
def json_entity_children(entity):
    """Yield JSON data for children of a ITree adapted entity.

    Each data will have `load_on_demand` unless it is a leaf of the tree.
    """
    for child in entity.cw_adapt_to('ITree').iterchildren():
        ctree = child.cw_adapt_to('ITree')
        on_demand = ctree and not ctree.is_leaf()
        ichild = child.cw_adapt_to('IJQTree')
        if ichild is not None:
            yield ichild.json(on_demand=on_demand)


@itree_adaptable
def json_entity_parents(entity):
    """Return the JSON data of the full tree walked upstream from `entity`.

    Each node along the path from this entity to the root will have
    `load_on_demand == False` so that the tree is opened down to `entity`.
    """
    # initialize data for "previous" parent when walking up the tree.
    previous, previous_children = None, []
    data = None
    for parent in entity.cw_adapt_to('ITree').iterparents():
        # parent has to be strictly evaluated
        iparent = parent.cw_adapt_to('IJQTree')
        if iparent is None:
            continue
        data = iparent.json(on_demand=False)
        data['children'] = []
        # fetch JSON for children of this parent, setting the previous
        # parent we're coming from as strict (load_on_demand=False).
        for child in json_entity_children(parent):
            if previous_children and child['id'] == previous:
                # remove load_on_demand as this is the branch we are
                # walking on.
                child.pop('load_on_demand', None)
                # add children of previous parent (which is a child of
                # this parent).
                child.setdefault('children', []).extend(previous_children)
            data['children'].append(child)
        # store this parent as the previous one for next iteration.
        previous, previous_children = parent.eid, data.get('children', [])
    return data


class JsonTreeView(json.JsonMixIn, EntityView):
    """JSON view for an entity adaptable as a ITree."""

    __regid__ = 'jqtree.json'
    __select__ = adaptable('ITree') & ~match_form_params('node')

    def entity_call(self, entity, **kwargs):
        # compute the parent JSON tree.
        data = json_entity_parents(entity)
        if not data:
            # if empty, just return the current entity JSON tree (with
            # children) as the entity is probably the root.
            adapted = entity.cw_adapt_to('IJQTree')
            if adapted is None:
                return
            data = adapted.json(on_demand=False)
            children = list(json_entity_children(entity))
            if children:
                data['children'] = children
        data['selected'] = entity.eid
        self.wdata([data])


class JsonTreeNodeView(json.JsonMixIn, View):
    """JSON view for a node of tree, returning the children of that node for
    lazy loading.
    """

    __regid__ = 'jqtree.json'
    __select__ = match_form_params('node')

    def call(self):
        eid = self._cw.form.pop('node')
        entity = self._cw.entity_from_eid(eid)
        children = list(json_entity_children(entity))
        self.wdata(children)


class JQTreeItemLabelView(EntityView):
    """View for the "label" of a node of a jqTree.

    Default to "oneline" view.
    """
    __regid__ = 'jqtree.label'

    def entity_call(self, entity):
        entity.view('oneline', w=self.w)


class JQTreeView(EntityView):
    """Tree view using jqTree library."""
    __regid__ = 'jqtree.treeview'
    __select__ = adaptable('ITree')

    def entity_call(self, entity, **kwargs):
        # TODO move jqTree js to its own file.
        self._cw.add_js(('cubes.jqtree.js', 'tree.jquery.js'))
        self._cw.add_css(('jqtree.css'))
        divid = 'jqtree' + str(entity.eid)
        data_url = entity.absolute_url(vid='jqtree.json')
        self.w(tags.div(id=divid, **{'data-url': data_url}))
        mayupdate = entity.e_schema.has_perm(
            self._cw, 'update', eid=entity.eid)
        self._cw.add_onload(uilib.js.jqtree.jqTree(divid, mayupdate))
